/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.integrationtest;

import javax.xml.ws.soap.SOAPFaultException;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.junit.Before;
import org.junit.Test;

import dk.nsi.dds.projects.sxa.testhelper.SxaLoggerHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaQueryHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaTestDataHelper;
import dk.nsi.dds.projects.sxa.testhelper.ddsinvoker.DDSInvokerHelper;
import dk.nsi.documentsharing.consumer.DocumentConsumer;
import dk.nsi.documentsharing.core.query.DocumentMetadataQueryResult;
import dk.nsi.documentsharing.core.query.FindDocumentsQuery;
import dk.nsi.documentsharing.core.retrieve.DocumentsToRetrieve;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocumentsResult;
import dk.nsi.documentsharing.testutil.DocumentSharingTestUtil;
import junit.framework.Assert;

/**
 * Tests performed by citizen type user
 * 
 * Tests of SXA documentsharing not affected by presence of other document metadata
 * providers or registered document metadata.
 * 
 * Please note that MappedCodedValue are used for specifying practiceSettingCode values where
 * CodedValue ought to be used. This was chosen to enable easier verification of actual outcome.
 */
public class CitizenQueryAndRetrieveIT {
  private static final Logger logger = Logger.getLogger(CitizenQueryAndRetrieveIT.class);
  private DocumentSharingTestUtil testutil = new DocumentSharingTestUtil();
  private static DocumentConsumer docConsumer;
  private static SxaQueryHelper queryHelper;
  private static SxaTestDataHelper testDataHelper;
  private static String userId = "1112132016";
  
  @Before
  public void setupClass() throws Exception {
    SxaLoggerHelper.setupLogger(Level.DEBUG); // Set to Level.DEBUG to print out
                                             // queries, returned metadata etc.
    testDataHelper = new SxaTestDataHelper();

    DDSInvokerHelper.resetCitizenInvocationContextProvider();
    docConsumer = DDSInvokerHelper.getCitizenConsumer(userId);
    queryHelper = new SxaQueryHelper(docConsumer, testDataHelper);
  }
  
  // ----------------------- Query Metadata Tests ------------------------------------

  /**
   * Query DocumentMetadata performed by the citizen him/herself. Expects a non-empty response
   */
  @Test
  public void metadataReturnedForQueryPerformedByCitizen() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getPatientId(userId));
    // Act
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertTrue(queryResult.getDocumentMetadataList().size() > 0);
  }
  
  /**
   * Query DocumentMetadata performed by the other citizen but with self relation. Expects SoapFaultException
   */
  @Test(expected = SOAPFaultException.class)
  public void metadataReturnedForQueryPerformedByOtherCitizen() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
    DDSInvokerHelper.getCitizenInvocationContextProvider().setSelfRelation();
    
    // Act
    docConsumer.doQuery(query);
  }

  /**
   * Query DocumentMetadata performed by parent of the the patient. Expects a non-empty response
   */
  @Test
  public void metadataReturnedForQueryPerformedByParent() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
    DDSInvokerHelper.getCitizenInvocationContextProvider().setParentRelation();
    
    // Act
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertTrue(queryResult.getDocumentMetadataList().size() > 0);
  }

  /**
   * Query DocumentMetadata performed by guardian of the the patient. Expects a non-empty response
   */
  @Test
  public void metadataReturnedForQueryPerformedByGuardian() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
    DDSInvokerHelper.getCitizenInvocationContextProvider().setGuardianRelation();
    
    // Act
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertTrue(queryResult.getDocumentMetadataList().size() > 0);
  }

  // ----------------------- Retrieve Document(s) Tests ------------------------------------

  /**
   * Retrieve document performed by patient's proxy holder. Expects a non-empty response
   */
  @Test
  public void retrieveDocumentsPerformedByProxyHolder() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
    DDSInvokerHelper.getCitizenInvocationContextProvider().setProxyHolderRelation();
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    // Act
    RetrievedDocumentsResult retrievedDocuments = docConsumer.doRetrieve(fetchList);
    // Assert
    logger.debug(testutil.prettyPrint(retrievedDocuments));
    Assert.assertTrue(retrievedDocuments.hasResult());
    Assert.assertNotNull(retrievedDocuments.getDocuments());
  }
  
  /**
   * Retrieve document performed by patient's parent. Expects a non-empty response
   */
  @Test
  public void retrieveDocumentsPerformedByParent() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
    DDSInvokerHelper.getCitizenInvocationContextProvider().setParentRelation();
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    // Act
    RetrievedDocumentsResult retrievedDocuments = docConsumer.doRetrieve(fetchList);
    // Assert
    logger.debug(testutil.prettyPrint(retrievedDocuments));
    Assert.assertTrue(retrievedDocuments.hasResult());
    Assert.assertNotNull(retrievedDocuments.getDocuments());
  }
  
  /**
   * Retrieve document performed by patient's guardian. Expects a non-empty response
   */
  @Test
  public void retrieveDocumentsPerformedByGuardian() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
    DDSInvokerHelper.getCitizenInvocationContextProvider().setGuardianRelation();
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    // Act
    RetrievedDocumentsResult retrievedDocuments = docConsumer.doRetrieve(fetchList);
    // Assert
    logger.debug(testutil.prettyPrint(retrievedDocuments));
    Assert.assertTrue(retrievedDocuments.hasResult());
    Assert.assertNotNull(retrievedDocuments.getDocuments());
  }
  
  /**
   * Retrieve document performed by the citizen him/herself. Expects a non-empty response
   */
  @Test
  public void retrieveDocumentsPerformedBySelf() throws Exception {
    // Arrange
    DDSInvokerHelper.resetCitizenInvocationContextProvider();
    docConsumer = DDSInvokerHelper.getCitizenConsumer(testDataHelper.getDefaultPatientId().getValue());
    
    queryHelper = new SxaQueryHelper(docConsumer, testDataHelper);
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
    DDSInvokerHelper.getCitizenInvocationContextProvider().setSelfRelation();
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    // Act
    RetrievedDocumentsResult retrievedDocuments = docConsumer.doRetrieve(fetchList);
    // Assert
    logger.debug(testutil.prettyPrint(retrievedDocuments));
    Assert.assertTrue(retrievedDocuments.hasResult());
    Assert.assertNotNull(retrievedDocuments.getDocuments());
  }
}
