/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.integrationtest;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.joda.time.DateTime;
import org.junit.BeforeClass;
import org.junit.Test;

import dk.nsi.dds.projects.sxa.common.MappedCodedValue;
import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaDocumentHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaLoggerHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaQueryHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaTestDataHelper;
import dk.nsi.dds.projects.sxa.testhelper.ddsinvoker.DDSInvokerHelper;
import dk.nsi.documentsharing.consumer.DocumentConsumer;
import dk.nsi.documentsharing.core.query.DocumentMetadataQueryResult;
import dk.nsi.documentsharing.core.query.FindDocumentsQuery;
import dk.nsi.documentsharing.core.query.SimpleDocumentMetadataQueryResult;
import dk.nsi.documentsharing.core.retrieve.DocumentsToRetrieve;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocumentsResult;
import dk.nsi.documentsharing.testutil.DocumentSharingTestUtil;
import junit.framework.Assert;

/**
 * Tests of SXA documentsharing not affected by presence of other document metadata
 * providers or registered document metadata.
 * 
 * Please note that MappedCodedValue are used for specifying practiceSettingCode values where
 * CodedValue ought to be used. This was chosen to enable easier verification of actual outcome.
 */
public class QueryAndRetrieveIT {
  private static final Logger logger = Logger.getLogger(QueryAndRetrieveIT.class);
  private DocumentSharingTestUtil testutil = new DocumentSharingTestUtil();
  private static DocumentConsumer docConsumer;
  private static SxaQueryHelper queryHelper;
  private static SxaTestDataHelper testDataHelper;
  private static TestDataHelper practiceSettingCodesDataHelper;
  
  @BeforeClass
  public static void setupClass() throws Exception {
    SxaLoggerHelper.setupLogger(Level.INFO); // Set to Level.DEBUG to print out
                                             // queries, returned metadata etc.
    docConsumer = DDSInvokerHelper.getConsumer();
    testDataHelper = new SxaTestDataHelper();
    queryHelper = new SxaQueryHelper(docConsumer, testDataHelper);
    practiceSettingCodesDataHelper = new TestDataHelper();
  }
  
  // ----------------------- Query Metadata Tests ------------------------------------

  /**
   * Query DocumentMetadata for an unknown patient. Expects an non-empty response
   * @throws Exception
   */
  @Test
  public void metadataReturnedForUnknownPatient() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getPatientIdForUnknownPatient());
    // Act
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertTrue(queryResult.getDocumentMetadataList().size() > 0);
  }
  
  @Test
  public void metadataReturnedForPatientHavingNoLabReports() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getPatientIdForPatientWithNoData());
    // Act
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertTrue(queryResult.getDocumentMetadataList().size() > 0);
  }
  
  /**
   * Verifies that a query: - using query parameters designating SXA-data - with list
   * of only valid practiceSettingCodes results in a single SXA document metadata
   * entry where: - practiceSettingCodes match the selected practiceSettingCodes
   */
  @Test
  public void metadataReturnedForMixedValidAndInvalidPracticeSettingCodes() throws Exception {
    // Arrange
    List<MappedCodedValue> mixedValidAndInvalidPracticeSettingCodes = new ArrayList<>();
    mixedValidAndInvalidPracticeSettingCodes.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForMicrobiology);
    mixedValidAndInvalidPracticeSettingCodes.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForPathology);
    mixedValidAndInvalidPracticeSettingCodes.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeNonSxa);
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(),
      mixedValidAndInvalidPracticeSettingCodes);
    // Act
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertNotNull(queryResult.getDocumentMetadataList());
    Assert.assertFalse(queryResult.getDocumentMetadataList().isEmpty());

    Assert.assertTrue(queryResult.hasDocumentMetadata());
  }
  
  /**
   * Query DocumentMetadata with a list of only invalid PracticeSettingCodes. Expects
   * an empty response
   * 
   * @throws Exception
   */
  @Test
  public void metadataReturnedForOnlyInvalidPracticeSettingCodes() throws Exception {
    // Arrange
    List<MappedCodedValue> selectedInvalidPracticeSettingCodes = new ArrayList<>();
    selectedInvalidPracticeSettingCodes.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeNonSxa);
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(),
      selectedInvalidPracticeSettingCodes);
    // Act
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertTrue(queryResult.getDocumentMetadataList().isEmpty()); 
  }
  
  /**
   * Verifies that a query: - using other query parameters than those designating
   * SXA-data results in no SXA document metadata entry.
   */
  @Test
  public void noSxaMetadataWhenQueryForTypeCodeNotUsedForSxa() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForNonSxaData();
    // Act
    SimpleDocumentMetadataQueryResult queryResult = docConsumer.doQueryForSimpleMetadata(query);
    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertTrue(queryResult.getDocumentMetadataList().isEmpty());
  }
  
  /**
   * Retrieve documents from a patient with a subset of possible
   * PracticeSettingCodes. Expects to retrieve only documents that match selected
   * PracticeSettingCodes
   * 
   * @throws Exception
   */
  
  // -------------------------- Retrieve Documents Tests -------------------------------
  
  @Test
  public void retrieveDocumentsWithSubsetOfPracticeSettingCodes() throws Exception {
    // Arrange
    List<MappedCodedValue> selectedValidPracticeSettingCodes = new ArrayList<>();
    selectedValidPracticeSettingCodes.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForMicrobiology);
    selectedValidPracticeSettingCodes.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForBiochemistry);
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(),
      selectedValidPracticeSettingCodes);
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    // Act
    RetrievedDocumentsResult retrievedDocuments = docConsumer.doRetrieve(fetchList);
    // Assert
    logger.debug(testutil.prettyPrint(retrievedDocuments));
    Assert.assertTrue(retrievedDocuments.hasResult());
    Assert.assertNotNull(retrievedDocuments.getDocuments());
    Assert.assertTrue(SxaDocumentHelper.ensureRetrievedDocumentsMatchSpecialities(selectedValidPracticeSettingCodes, retrievedDocuments));
  }

  /**
   * Retrieve documents from a patient with all possible PracticeSettingCodes.
   * Expects to retrieve all documents from patient
   * 
   * @throws Exception
   */
  @Test
  public void retrieveDocumentsWithAllPracticeSettingCodes() throws Exception {
    // Arrange
    List<MappedCodedValue> allPracticeSettingCodes = new ArrayList<>();
    allPracticeSettingCodes.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForMicrobiology);
    allPracticeSettingCodes.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForBiochemistry);
    allPracticeSettingCodes.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForPathology);
    
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), allPracticeSettingCodes);
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    // Act
    RetrievedDocumentsResult retrievedDocuments = docConsumer.doRetrieve(fetchList);
    // Assert
    logger.debug(testutil.prettyPrint(retrievedDocuments));
    Assert.assertTrue(retrievedDocuments.hasResult());
    Assert.assertTrue(SxaDocumentHelper.ensureRetrievedDocumentsMatchSpecialities(allPracticeSettingCodes, retrievedDocuments));
  }

  /**
   * Retrieve documents from a patient with single valid PracticeSettingCode.
   * Expects to retrieve documents for patient
   * 
   * @throws Exception
   */
  @Test
  public void retrieveDocumentsWithSinglePracticeSettingCode() throws Exception {
    // Arrange
    List<MappedCodedValue> allPracticeSettingCodes = new ArrayList<>();
    allPracticeSettingCodes.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForMicrobiology);
    
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), allPracticeSettingCodes);
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    // Act
    RetrievedDocumentsResult retrievedDocuments = docConsumer.doRetrieve(fetchList);
    // Assert
    logger.debug(testutil.prettyPrint(retrievedDocuments));
    Assert.assertTrue(retrievedDocuments.hasResult());
    Assert.assertTrue(SxaDocumentHelper.ensureRetrievedDocumentsMatchSpecialities(allPracticeSettingCodes, retrievedDocuments));
  }
  
  /**
   * Retrieve documents from a patient with no PracticeSettingCodes.
   * Expects to retrieve all documents from patient
   * 
   * @throws Exception
   */
  @Test
  public void retrieveDocumentsWithNoPracticeSettingCodes() throws Exception {
    
    // Arrange
    List<MappedCodedValue> emptyList = new ArrayList<>();
    
    List<MappedCodedValue> fullList = new ArrayList<>();
    fullList.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForMicrobiology);
    fullList.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForBiochemistry);
    fullList.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForPathology);
    
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), emptyList);
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    
    // Act
    RetrievedDocumentsResult retrievedDocumentsResult = docConsumer.doRetrieve(fetchList);
    
    // Assert
    Assert.assertTrue(retrievedDocumentsResult.hasResult());
    Assert.assertTrue(SxaDocumentHelper.ensureRetrievedDocumentsMatchSpecialities(fullList, retrievedDocumentsResult));
  }
  /**
   * Retrieve documents from a patient with both valid sxa practice setting codes and invalid sxa practice setting codes.
   * Expects to retrieve only documents that match the valid practice setting codes
   * 
   * @throws Exception
   */
  @Test
  public void retrieveDocumentsWithMixedValidAndInvalidPracticeSettingCodes() throws Exception {
    
    // Arrange
    List<MappedCodedValue> practiceSettingCodesWithInvalid = new ArrayList<>();
    practiceSettingCodesWithInvalid.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForBiochemistry);
    practiceSettingCodesWithInvalid.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeNonSxa);
    
    List<MappedCodedValue> validList = new ArrayList<>();
    validList.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForBiochemistry);
    
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), practiceSettingCodesWithInvalid);
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    
    // Act
    RetrievedDocumentsResult retrievedDocumentsResult = docConsumer.doRetrieve(fetchList);
    
    // Assert
    Assert.assertTrue(retrievedDocumentsResult.hasResult());
    Assert.assertTrue(SxaDocumentHelper.ensureRetrievedDocumentsMatchSpecialities(validList, retrievedDocumentsResult));
  }
  
  /**
   * Retrieve documents from a patient with only invalid PracticeSettingCodes.
   * Expects to retrieve response with no documents 
   * 
   * @throws Exception
   */
  @Test
  public void retrieveDocumentsWithOnlyInvalidPracticeSettingCodes() throws Exception {
    
    // Arrange
    List<MappedCodedValue> invalidList = new ArrayList<>();
    invalidList.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeNonSxa);
       
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(testDataHelper.getDefaultPatientId(), invalidList);
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);

    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertTrue(queryResult.getDocumentMetadataList().isEmpty());
  }
  
  /**
   * Retrieve documents from a patient with service start and stop time set.
   * Expects to retrieve response
   * 
   * @throws Exception
   */
  @Test
  public void retrieveDocumentsWithStartAndStopTime() throws Exception {
    
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQuery(testDataHelper.getDefaultPatientId(), DateTime.now(), DateTime.now());
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    
    // Act
    RetrievedDocumentsResult retrievedDocumentsResult = docConsumer.doRetrieve(fetchList);

    // Assert
    Assert.assertTrue(retrievedDocumentsResult.hasResult());
  }
  
  /**
   * Retrieve documents from a patient with service start time explicitly set and stop time set to default value.
   * Expects to retrieve response
   * 
   * @throws Exception
   */
  @Test
  public void retrieveDocumentsWithStartTime() throws Exception {
    
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQuery(testDataHelper.getDefaultPatientId(), DateTime.now(), null);
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    
    // Act
    RetrievedDocumentsResult retrievedDocumentsResult = docConsumer.doRetrieve(fetchList);

    // Assert
    Assert.assertTrue(retrievedDocumentsResult.hasResult());
  }
  
  /**
   * Retrieve documents from a patient with service stop time explicitly set and start time set to default value.
   * Expects to retrieve response
   * 
   * @throws Exception
   */
  @Test
  public void retrieveDocumentsWithStopTime() throws Exception {
    
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQuery(testDataHelper.getDefaultPatientId(), null, DateTime.now());
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    
    // Act
    RetrievedDocumentsResult retrievedDocumentsResult = docConsumer.doRetrieve(fetchList);

    // Assert
    Assert.assertTrue(retrievedDocumentsResult.hasResult());
  }
  
  /**
   * Retrieve documents from a patient with service start and stop time set to.
   * Expects to retrieve response
   * 
   * @throws Exception
   */
  @Test
  public void retrieveDocumentsWithoutStartAndStopTime() throws Exception {
    
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQuery(testDataHelper.getDefaultPatientId(), null, null);
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    
    // Act
    RetrievedDocumentsResult retrievedDocumentsResult = docConsumer.doRetrieve(fetchList);

    // Assert
    Assert.assertTrue(retrievedDocumentsResult.hasResult());
  }
}
