/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.integrationtest;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;

import dk.nsi.dds.projects.sxa.common.MappedCodedValue;
import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaConsentHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaLoggerHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaQueryHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaTestDataHelper;
import dk.nsi.dds.projects.sxa.testhelper.ddsinvoker.DDSInvokerHelper;
import dk.nsi.documentsharing.consent.helper.ConsentHelper;
import dk.nsi.documentsharing.consumer.DocumentConsumer;
import dk.nsi.documentsharing.consumer.helper.DocumentSharingHelper;
import dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthority;
import dk.nsi.documentsharing.core.query.DocumentMetadataQueryResult;
import dk.nsi.documentsharing.core.query.FindDocumentsQuery;
import dk.nsi.documentsharing.core.retrieve.DocumentsToRetrieve;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocumentsResult;
import dk.nsi.documentsharing.docsharingservices.context.DSSInvocationContextProvider;
import dk.nsi.documentsharing.nationalprofile.NationalProfileValueFactory;
import dk.nsi.documentsharing.testutil.DocumentSharingTestUtil;
import junit.framework.Assert;

/**
 * 
 * Test procedure:
 * BeforeClass:
 *  retrieve all SXA documents using break-the-glass
 * Testcases:
 *  Setup data specific consent based on test case and previous retrieved documents
 *  Verify DDS data specific consent behavior 
 *
 */

public class SxaConsentIT {

  private static final Logger logger = Logger.getLogger(SxaConsentIT.class);
  private DocumentSharingTestUtil testutil = new DocumentSharingTestUtil();
  private static DocumentConsumer docConsumer; 
  private static SxaQueryHelper queryHelper;
  private static DSSInvocationContextProvider invocationContextProvider;
  private static NationalProfileValueFactory valueFactory = new NationalProfileValueFactory();
  private static SxaConsentHelper consentHelper;
  private static TestDataHelper practiceSettingCodesDataHelper = new TestDataHelper();

  @BeforeClass
  public static void setup() throws Exception {
    invocationContextProvider = DDSInvokerHelper.getInvocationContextProvider();
    consentHelper = new SxaConsentHelper(new ConsentHelper(
      DocumentSharingHelper.getConfigurationProvider(),
      invocationContextProvider,
      valueFactory));

    SxaLoggerHelper.setupLogger(Level.DEBUG); // Set to Level.DEBUG to print out queries, returned metadata etc.
    docConsumer = DDSInvokerHelper.getConsumer();
    queryHelper = new SxaQueryHelper(docConsumer, new SxaTestDataHelper());
  }

  @AfterClass
  public static void tearDownAll() throws Exception {
    consentHelper.revokeAllConsents(getPatientId());
  }
  
  @Test
  public void didRemoveResponseOverlappingConsent() throws Exception {
    // Arrange
    String patientCPR = getPatientId();
    ValueWithAssigningAuthority patientId = valueFactory.createPatientIdUsingCpr(patientCPR);

    String organisationProviderNumber = "6008319";
    String organisationSorCode = "247081000016006"; // Found manually as corresponding to organisationProviderNumber
    
    List<MappedCodedValue> selectedValidPracticeSettingCodes = new ArrayList<>();
    selectedValidPracticeSettingCodes.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForMicrobiology);
    selectedValidPracticeSettingCodes.add(practiceSettingCodesDataHelper.mappedPracticeSettingCodeForBiochemistry);
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData(patientId,
      selectedValidPracticeSettingCodes);
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());

    RetrievedDocumentsResult retrievedDocumentsWithNoConsents = docConsumer.doRetrieve(fetchList);
    consentHelper.ensureNegativeConsentForPatientAgainstInformationFrom(patientCPR, organisationSorCode);
  
    // Act
    RetrievedDocumentsResult retrievedDocumentsWithConsents = docConsumer.doRetrieve(fetchList);
    
    // Assert
    int numberOfDocumentsBefore = consentHelper.getNumberOfDocuments(retrievedDocumentsWithNoConsents);
    int numberOfDocumentsAfter = consentHelper.getNumberOfDocuments(retrievedDocumentsWithConsents);
    
    System.out.println("--------------- documents: " + numberOfDocumentsBefore + ", " + numberOfDocumentsAfter);
    
    Assert.assertTrue(numberOfDocumentsAfter < numberOfDocumentsBefore);
    Assert.assertTrue(retrievedDocumentsWithConsents.isFilteredByConsent());
  }
  private static String getPatientId() {
    return SxaTestDataHelper.DEFAULT_PATIENT_CPR;
  }
}
