/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.integrationtest;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.joda.time.DateTime;
import org.joda.time.Interval;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;

import dk.nsi.dds.projects.sxa.testhelper.SxaDocumentHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaLoggerHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaQueryHelper;
import dk.nsi.dds.projects.sxa.testhelper.SxaTestDataHelper;
import dk.nsi.dds.projects.sxa.testhelper.ddsinvoker.DDSInvokerHelper;
import dk.nsi.documentsharing.consumer.DocumentConsumer;
import dk.nsi.documentsharing.core.query.DocumentMetadataQueryResult;
import dk.nsi.documentsharing.core.query.FindDocumentsQuery;
import dk.nsi.documentsharing.core.retrieve.DocumentsToRetrieve;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocumentsResult;
import dk.nsi.documentsharing.testutil.DocumentSharingTestUtil;
import dk.nsi.documentsharing.xds.impl.ValueConverter;
import junit.framework.Assert;

/**
 * These tests assume that SXA is the only document metadata provider responding to
 * queries. Presence of other document metadata providers or document metadata
 * registered by such in a shared registry are likely to interfere with these tests.
 */
@Ignore
public class SxaExclusiveIT {
  private static final Logger logger = Logger.getLogger(SxaExclusiveIT.class);
  private DocumentSharingTestUtil testutil = new DocumentSharingTestUtil();
  private static DocumentConsumer docConsumer;
  private static SxaQueryHelper queryHelper;
  private static SxaTestDataHelper testDataHelper;
  private static SxaDocumentHelper documentHelper;
  private static DateTime start = ValueConverter.convertFromDTM("20140127" + "104056");
  private static DateTime stop = start.plusYears(1);

  @BeforeClass
  public static void setupClass() throws Exception {
    SxaLoggerHelper.setupLogger(Level.DEBUG); // Set to Level.DEBUG to print out
                                              // queries, returned metadata etc.
    docConsumer = DDSInvokerHelper.getConsumer();
    queryHelper = new SxaQueryHelper(docConsumer, new SxaTestDataHelper());
    testDataHelper = new SxaTestDataHelper();
    documentHelper = new SxaDocumentHelper();
  }

  /**
   * Verifies that a query: - using query parameters designating SXA-data - without
   * service start and stop query parameters specified results in a single SXA
   * document metadata entry.
   */
  @Test
  public void sxaMetadataReturnedWithoutStartAndStop() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData();
    logger.debug(testutil.prettyPrint(query));
    // Act
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    logger.debug(testutil.prettyPrint(queryResult));
    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertEquals(1, queryResult.getDocumentMetadataList().size());
  }

  /**
   * Verifies that a query: - using query parameters designating SXA-data - with
   * service start and stop query parameters specified results in a single SXA
   * document metadata entry where: - serviceStartTime is set to start time of the
   * query - serviceStopTime is set to stop time of the query
   */
  @Test
  public void sxaMetadataReturnedWithServiceStartAndStop() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData().setServiceStartTime(start).setServiceStopTime(stop);
    logger.debug(testutil.prettyPrint(query));
    // Act
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    logger.debug(testutil.prettyPrint(queryResult));
    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertEquals(start,
      queryResult.getDocumentMetadataList().get(0).getServiceStartTime().withZone(start.getZone()));
    Assert.assertEquals(stop,
      queryResult.getDocumentMetadataList().get(0).getServiceStopTime().withZone(stop.getZone()));
  }

  /**
   * Verifies that a query: - using query parameters designating SXA-data - with
   * service start interval query parameters specified results in a single SXA
   * document metadata entry where: - serviceStartTime is set to the beginning of the
   * start interval of the query - serviceStopTime is set to null
   */
  @Test
  public void sxaMetadataReturnedWithServiceStartInterval() throws Exception {
    // Arrange
    DateTime startBegin = start;
    DateTime startEnd = stop;
    Interval interval = new Interval(startBegin, startEnd);
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData();
    query = query.setServiceStartTime(interval);
    logger.debug(testutil.prettyPrint(query));
    // Act
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    logger.debug(testutil.prettyPrint(queryResult));
    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertEquals(1, queryResult.getDocumentMetadataList().size());
    Assert.assertEquals(startBegin,
      queryResult.getDocumentMetadataList().get(0).getServiceStartTime().withZone(startBegin.getZone()));
    Assert.assertNull(queryResult.getDocumentMetadataList().get(0).getServiceStopTime());
  }

  /**
   * Verifies that a query: - using query parameters designating SXA-data - with
   * service stop query parameter specified results in a single SXA document metadata
   * entry where: - serviceStartTime is set to null - serviceStopTime is set to the
   * stop value of the query
   */
  @Test
  public void sxaMetadataReturnedWithServiceStop() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData().setServiceStopTime(stop);
    logger.debug(testutil.prettyPrint(query));
    // Act
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    logger.debug(testutil.prettyPrint(queryResult));
    // Assert
    Assert.assertTrue(queryResult.hasResult());
    Assert.assertNull(queryResult.getDocumentMetadataList().get(0).getServiceStartTime());
    Assert.assertEquals(stop,
      queryResult.getDocumentMetadataList().get(0).getServiceStopTime().withZone(stop.getZone()));
  }

  /**
   * Verifies that a query: - using query parameters designating SXA-data - with
   * service start and stop query parameters specified and subsequent document
   * retrieval results in a single SXA document
   */
  @Test
  public void sxaReturnedWithStartAndStop() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData().setServiceStartTime(start).setServiceStopTime(stop);
    logger.debug(testutil.prettyPrint(query));
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    logger.debug(testutil.prettyPrint(queryResult));
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    // Act
    RetrievedDocumentsResult retrievedDocuments = docConsumer.doRetrieve(fetchList);
    logger.debug(testutil.prettyPrint(retrievedDocuments));
    // Assert
    Assert.assertTrue(retrievedDocuments.hasResult());
    Assert.assertEquals(1, retrievedDocuments.getDocuments().size());
  }

  /**
   * Verifies that a query: - using query parameters designating SXA-data - with
   * service start query parameter specified and subsequent document retrieval
   * results in a single SXA document
   */
  @Test
  public void sxaDocumentReturnedWithStart() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData().setServiceStartTime(start);
    logger.debug(testutil.prettyPrint(query));
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    logger.debug(testutil.prettyPrint(queryResult));
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    // Act
    RetrievedDocumentsResult retrievedDocuments = docConsumer.doRetrieve(fetchList);
    logger.debug(testutil.prettyPrint(retrievedDocuments));
    // Assert
    Assert.assertTrue(retrievedDocuments.hasResult());
    Assert.assertEquals(1, retrievedDocuments.getDocuments().size());
  }

  /**
   * Verifies that a query: - using query parameters designating SXA-data - with
   * service stop query parameter specified and subsequent document retrieval results
   * in a single SXA document
   */
  @Test
  public void sxaDocumentReturnedWithStop() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData().setServiceStopTime(stop);
    logger.debug(testutil.prettyPrint(query));
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    logger.debug(testutil.prettyPrint(queryResult));
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    // Act
    RetrievedDocumentsResult retrievedDocuments = docConsumer.doRetrieve(fetchList);
    logger.debug(testutil.prettyPrint(retrievedDocuments));
    // Assert
    Assert.assertTrue(retrievedDocuments.hasResult());
    Assert.assertEquals(1, retrievedDocuments.getDocuments().size());
  }

  /**
   * Verifies that a query: - using query parameters designating SXA-data - without
   * start or stop query parameter specified and subsequent document retrieval
   * results in a single SXA document
   */
  @Test
  public void sxaDocumentReturnedWithoutStartAndStop() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData();
    logger.debug(testutil.prettyPrint(query));
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    logger.debug(testutil.prettyPrint(queryResult));
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    // Act
    RetrievedDocumentsResult retrievedDocuments = docConsumer.doRetrieve(fetchList);
    logger.debug(testutil.prettyPrint(retrievedDocuments));
    // Assert
    Assert.assertTrue(retrievedDocuments.hasResult());
    Assert.assertEquals(1, retrievedDocuments.getDocuments().size());
  }

  /**
   * Verifies that a query: - using query parameters designating SXA-data - with
   * (past) start and future stop (one year from now) and subsequent document
   * retrieval results in a single SXA document
   */
  @Test
  public void sxaDocumentReturnedForFutureStop() throws Exception {
    // Arrange
    FindDocumentsQuery query = queryHelper.prepareQueryForSxaData()
      .setServiceStartTime(DateTime.now().minusYears(1))
      .setServiceStopTime(DateTime.now().plusYears(1));
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    DocumentsToRetrieve fetchList = docConsumer.prepareRetrieve(queryResult.getDocumentMetadataList());
    // Act
    RetrievedDocumentsResult retrievedDocuments = docConsumer.doRetrieve(fetchList);
    // Assert
    Assert.assertTrue(retrievedDocuments.hasResult());
  }
}
