/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.testhelper;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;
import org.joda.time.Interval;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import ca.uhn.hl7v2.model.DataTypeException;
import ca.uhn.hl7v2.model.primitive.CommonTS;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocument;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocumentsResult;
import dk.nsi.documentsharing.nationalprofile.NationalProfileConstants;
import junit.framework.Assert;

public class SxaConsentDocumentHelper {
  
  public static boolean hasNoReports(RetrievedDocumentsResult retrievedDocuments) throws Exception {
    return countEntriesInSectionTreatmentPlan(XPathFactory.newInstance().newXPath(), getDOM(retrievedDocuments)) == 0;
  }

  public static boolean hasReportContent(RetrievedDocument retrievedDocument) throws Exception {
    return countEntriesInSectionTreatmentPlan(XPathFactory.newInstance().newXPath(), getDOM(retrievedDocument)) > 0;
  }
  
  public static Interval getMaximumSpanOfReports(RetrievedDocument retrievedDocument) 
      throws UnsupportedEncodingException, ParserConfigurationException, SAXException, IOException, XPathExpressionException {
    return getServiceEventEffectiveTime(XPathFactory.newInstance().newXPath(), getDOM(retrievedDocument));
  }

  public static List<String> getReportOrganisations(RetrievedDocument retrievedDocument) 
      throws UnsupportedEncodingException, ParserConfigurationException, SAXException, IOException, XPathExpressionException {
    return Stream.of(getAuthorAssignedAuthorSorCode(XPathFactory.newInstance().newXPath(), getDOM(retrievedDocument))).collect(Collectors.toList());
  }

  private static Document getDOM(RetrievedDocumentsResult retrievedDocuments) 
      throws UnsupportedEncodingException, ParserConfigurationException, SAXException, IOException {
    return getDOM(retrievedDocuments.getDocuments().get(0));
  }

  private static Document getDOM(RetrievedDocument retrievedDocument)
      throws ParserConfigurationException, SAXException, IOException,
      UnsupportedEncodingException {
    return SxaXmlUtil.parseXmltextToDocument(new String(retrievedDocument.getDocument(), "UTF-8"));
  }

  private static String getAuthorAssignedAuthorSorCode(XPath xpath, Document document) throws XPathExpressionException {
    XPathExpression expr = xpath.compile("/ClinicalDocument/author/assignedAuthor/id"); // TODO: FIX xml path
    try {
      Element element = (Element) performXPath(expr, document);
      String rootValue = element.getAttribute("root");
      if (Arrays.asList(NationalProfileConstants.ASSIGNINGAUTHORITY_OID_DANISH_ORGANISATION_SOR_ID, NationalProfileConstants.ASSIGNINGAUTHORITY_OID_DANISH_ORGANISATION_SOR_ID + ".1").contains(rootValue)) {
        return element.getAttribute("extension");
      }
    } catch (XPathExpressionException xe) {
      Assert.fail("Document has entries in Treatment Plan section when it should not");
    }
    return null;
  }

  private static Interval getServiceEventEffectiveTime(XPath xpath, Document document) throws XPathExpressionException {
    Interval result = null;
    XPathExpression lowExpr = xpath.compile("/ClinicalDocument/documentationOf/serviceEvent/effectiveTime/low");
    XPathExpression hiExpr = xpath.compile("/ClinicalDocument/documentationOf/serviceEvent/effectiveTime/high");
    try {
      Element element = (Element) performXPath(lowExpr, document);
      String lowValue = element.getAttribute("value");
      element = (Element) performXPath(hiExpr, document);
      String highValue = element.getAttribute("value");
      result = new Interval(fromCDA(lowValue), fromCDA(highValue));
    } catch (XPathExpressionException xe) {
      Assert.fail("Document has entries in Treatment Plan section when it should not");
    }
    return result;
  }

  public static DateTime fromCDA(String timestamp) {
    if (timestamp == null) {
      return null;
    }
    if ((timestamp.indexOf('-') < 0) && (timestamp.indexOf('+') < 0)) {
      timestamp += "+0000";
    }

    try {
      CommonTS ts = new CommonTS(timestamp);
      DateTime dateTime = new DateTime(
        ts.getYear(),
        (ts.getMonth() == 0) ? 1 : ts.getMonth(),
          (ts.getDay() == 0) ? 1 : ts.getDay(),
            ts.getHour(),
            ts.getMinute(),
            ts.getSecond(),
            DateTimeZone.forOffsetHoursMinutes(
              ts.getGMTOffset() / 100,
              ts.getGMTOffset() % 100));
      return dateTime.toDateTime(DateTimeZone.UTC);
    } catch (DataTypeException e) {
      throw new IllegalArgumentException("Invalid timestamp " + timestamp);
    }
  }  

  private static long countEntriesInSectionTreatmentPlan(XPath xpath, Document document) throws XPathExpressionException {
    XPathExpression expr = xpath.compile("count(/ClinicalDocument/component/structuredBody/component/section[//@code='18776-5']/entry)");
    try {
      String result = expr.evaluate(document);
      Assert.assertNotNull(result);
      return Long.valueOf(result);
    } catch (XPathExpressionException xe) {
      Assert.fail("Failed to count number of entries in Treatment Plan section");
    }
    return 0;
  }
  
  private static Element performXPath(XPathExpression xpathExpression, Document document) throws XPathExpressionException {
    return (Element) xpathExpression.evaluate(document, XPathConstants.NODE);
  }
}
