/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.testhelper;

import java.io.StringReader;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathFactory;

import org.bouncycastle.util.encoders.Base64;
import org.omg.CosNaming._BindingIteratorImplBase;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import dk.nsi.dds.projects.sxa.common.MappedCodedValue;
import dk.nsi.documentsharing.core.metadata.model.CodedValue;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocumentsResult;

public class SxaDocumentHelper {
  
  public static final String mappedXstring = "/Rekvisition/LaboratorieSpeciale";
  public static final String rawXstring = "/*[local-name()='LaboratoryReport']/*[local-name()='Letter']/*[local-name()='TypeCode']";
  
  public static boolean ensureRetrievedDocumentsMatchSpecialities(List<MappedCodedValue> inList, RetrievedDocumentsResult retrievedDocumentsResult) {
    try {
      byte[] document = retrievedDocumentsResult.getDocuments().get(0).getDocument(); 
      List<String> mappedDocuments = getDocumentsArrayFromRetrievedDocument(document, "/ResultatPakkePIH/MappedeSvar/DocumentWrapper");
      boolean mappedDocumentsMatch = ensureDocumentsMatchSpecialities(inList, mappedDocuments, mappedXstring, false) && ensureAllSpecialitiesIsInListOfDocuments(inList, mappedDocuments, mappedXstring, false);
      
      List<String> rawDocuments = getDocumentsArrayFromRetrievedDocument(document, "/ResultatPakkePIH/RaaSvar/LeverandoerSvar/LeverandoerSvarXml/DocumentWrapper");
      boolean rawDocumentsMatch = ensureDocumentsMatchSpecialities(inList, rawDocuments, rawXstring, true) && ensureAllSpecialitiesIsInListOfDocuments(inList, rawDocuments, rawXstring, true);
    
      return mappedDocumentsMatch && rawDocumentsMatch;
    } catch (Exception e) {
      e.printStackTrace();
    }
   
    return false;
  }
  
  private static boolean ensureDocumentsMatchSpecialities(List<MappedCodedValue> inList, List<String> documents, String xString, boolean xrptType) {
    for (String document : documents) {
      if (!documentContainsValidSpeciality(document, inList, xString, xrptType)) {
        return false;
      }
    }
    return true;
  }
  
  private static boolean ensureAllSpecialitiesIsInListOfDocuments(List<MappedCodedValue> inList, List<String> documents, String xString, boolean xrptType) {
    for (MappedCodedValue speciality : inList) {
      if (!specialityIsInDocuments(speciality, documents, xString, xrptType)) {
        return false;
      }
    }
    return true;
  }
  
  private static boolean specialityIsInDocuments(MappedCodedValue speciality, List<String> documents, String xString, boolean xrptType) {
    
    for (String doc : documents) {
      String specialityInDocument;
      try {
        specialityInDocument = findSpeciality(doc, xString);
        
        if (xrptType) {
          if (speciality.getMappedValue().equals(convertXrptToMappedValue(specialityInDocument))) {
            return true;
          }
        } else {
          if (speciality.getMappedValue().equals(specialityInDocument)) {
            return true;
          }
        }

      } catch (Exception e) {
        e.printStackTrace();
      }
    }
    return false;
  }
  
  private static boolean documentContainsValidSpeciality(String document, List<MappedCodedValue> specialities, String xString, boolean xrptType) { 
    try {
      String specialityInDocument = findSpeciality(document, xString);
      for (MappedCodedValue codedValue : specialities) {
        if (xrptType) {
          if (codedValue.getMappedValue().equals(convertXrptToMappedValue(specialityInDocument))) {
            return true;
          }
        } else {
          if (codedValue.getMappedValue().equals(specialityInDocument)) {
            return true;
          }
        }
      }
    } catch (Exception e) {
      // TODO: handle exception
    }
    
    return false;
  }
  
  private static List<String> getDocumentsArrayFromRetrievedDocument(byte[] document, String xString) throws Exception {
    List<String> resultArray = new ArrayList<String>();
    
    String xml = new String(document, "utf-8");
    String trimmedXml = xml.trim().replaceFirst("^([\\W]+)<","<");
    
    DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
    DocumentBuilder builder = factory.newDocumentBuilder();
    Document doc = builder.parse(new InputSource(new StringReader(trimmedXml)));
    XPathFactory xPathfactory = XPathFactory.newInstance();
    XPath xpath = xPathfactory.newXPath();
    XPathExpression expr = xpath.compile(xString);
    NodeList nodeList = (NodeList) expr.evaluate(doc, XPathConstants.NODESET);
    
    for (int i = 0; i < nodeList.getLength(); i++) {
      Node node = nodeList.item(i);
      resultArray.add(node.getFirstChild().getNodeValue());
    }
 
    return resultArray;
  }
  
  private static String findSpeciality(String doc, String xString) throws Exception {
    byte[] raw = Base64.decode(doc.getBytes());
    
    String xml = new String(raw, "utf-8");
    String trimmedXml = xml.trim().replaceFirst("^([\\W]+)<","<");
    XPath xPath = XPathFactory.newInstance().newXPath();
    String result = xPath.evaluate(xString, new InputSource(new StringReader(trimmedXml)));
  
    return result;
  }
  
  private static String convertXrptToMappedValue(String xprt) {
    switch (xprt) {
      case "XRPT01":
        return "KliniskBiokemi";
      case "XRPT02":
        return "Mikrobiologi";
      case "XRPT04":
        return "Patologi";
      default:
        return "";
    }
  }
}
