/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.testhelper;

import java.util.List;

import org.joda.time.DateTime;

import dk.nsi.dds.projects.sxa.common.MappedCodedValue;
import dk.nsi.documentsharing.consumer.DocumentConsumer;
import dk.nsi.documentsharing.core.metadata.model.CodedValue;
import dk.nsi.documentsharing.core.metadata.model.CodedValueImpl;
import dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthority;
import dk.nsi.documentsharing.core.query.FindDocumentsQuery;

public class SxaQueryHelper {

	public static final String PATIENT_CPR = "2512484916";
	public static final String UNKNOWN_PATIENT_CPR = "0101010101";
	private static CodedValue sxaTypeCode = new CodedValueImpl("11502-2", "2.16.840.1.113883.6.1", "LABORATORY REPORT.TOTAL");
	private static CodedValue nonsxaTypeCode = new CodedValueImpl("53576-5", "2.16.840.1.113883.6.1", "Personal Health Monitoring Report");
	private static CodedValue sxaFormatCode = new CodedValueImpl("urn:ad:dk:medcom:labreports:svareksponeringsservice", "1.2.208.184.100.10", "Laboratoriesvar (samling af svar)");

	private DocumentConsumer docConsumer;
	private SxaTestDataHelper testDataHelper;

	public SxaQueryHelper(DocumentConsumer docConsumer, SxaTestDataHelper testDataHelper) {
		super();
		this.docConsumer = docConsumer;
		this.testDataHelper = testDataHelper;
	}

	public FindDocumentsQuery prepareQueryForSxaData() {
		return prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
	}

	public FindDocumentsQuery prepareQueryForSxaData(ValueWithAssigningAuthority patientId) {
		return prepareQuery(patientId, sxaTypeCode, sxaFormatCode);
	}

	public FindDocumentsQuery prepareQueryForNonSxaData() {
		return prepareQuery(testDataHelper.getDefaultPatientId(), nonsxaTypeCode, sxaFormatCode);
	}
	
	public FindDocumentsQuery prepareQueryForSxaData(ValueWithAssigningAuthority patientId, List<MappedCodedValue> practiceSettingCodes) {
	  FindDocumentsQuery query = prepareQuery(patientId, sxaTypeCode, sxaFormatCode);
	  for(CodedValue code : practiceSettingCodes) {
	    query.addPracticeSettingCode(code);
	  }
	  return query;
	}
	
	public FindDocumentsQuery prepareQueryForSxaDataWithOnlyUnsupportedPracticeSettingCodes() {
      FindDocumentsQuery query = prepareQueryForSxaData(testDataHelper.getDefaultPatientId());
      query.addPracticeSettingCode(testDataHelper.practiceSettingCodeInvalid);
      return query;
    }

	private FindDocumentsQuery prepareQuery(ValueWithAssigningAuthority patientId, CodedValue typeCode, CodedValue formatCode) {
		return prepareQuery(patientId, typeCode, formatCode, DateTime.now(), DateTime.now());
	}

	private FindDocumentsQuery prepareQuery(ValueWithAssigningAuthority patientId,
	  CodedValue typeCode, CodedValue formatCode,
	  DateTime startTime, DateTime stopTime) {
	  FindDocumentsQuery query = docConsumer.createQuery(patientId).
	      addQueryForOnDemandDocuments().
	      addTypeCode(typeCode).
	      addFormatCode(formatCode);
	  if (startTime !=null) {
	    query.setServiceStartTime(startTime);
	  }
	  if (stopTime !=null) {
	    query.setServiceStopTime(stopTime);
	  }
	  return query;
	}

	public FindDocumentsQuery prepareQuery(ValueWithAssigningAuthority patientId, DateTime startTime, DateTime stopTime) {
	  return prepareQuery(patientId, sxaTypeCode, sxaFormatCode, startTime, stopTime);
	}

}
