/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2014 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.testhelper.ddsinvoker;

import java.util.Properties;

import javax.xml.ws.Holder;

import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;
import org.w3c.dom.Document;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthority;
import dk.nsi.documentsharing.docsharingservices.context.DSSInvocationContext;
import dk.nsi.documentsharing.docsharingservices.context.DSSSystemContextProvider;
import dk.nsi.hsuid.CitizenUserRelationAttribute.LegalValues;
import dk.nsi.hsuid.consumer.CitizenHsuid;
import dk.nsi.hsuid.consumer.HsuidConsumer;
import dk.nsi.hsuid.consumer.impl.HsuidConsumerImpl;
import dk.nsi.service.HeaderConverter;
import dk.sosi.seal.SOSIFactory;
import dk.sosi.seal.model.IDCard;
import dk.sosi.seal.model.Request;
import dk.sosi.seal.model.SignatureUtil;
import dk.sosi.seal.vault.EmptyCredentialVault;

public class DSSCitizenInvocationContextProviderImpl implements DSSCitizenInvocationContextProvider {

  private static final String HSUID_HEADER_ISSUER = "SXA DSSDocumentConsumer";

  private DSSSystemContextProvider systemContextProvider;
  private DSSCitizenUserContextProvider userContextProvider;
  private HsuidConsumer hsuidConsumer;
  private String flowId;
  private LegalValues relation;

  public DSSCitizenInvocationContextProviderImpl(DSSSystemContextProvider systemContextProvider, DSSCitizenUserContextProvider userContextProvider) {
    this.systemContextProvider = systemContextProvider;
    this.userContextProvider = userContextProvider;
    this.hsuidConsumer = new HsuidConsumerImpl(
      systemContextProvider.getSystemOwnerName(),
      systemContextProvider.getSystemName(),
      systemContextProvider.getSystemVersion(),
      systemContextProvider.getOperationsOrganisationName(),
      HSUID_HEADER_ISSUER);
  }

  /* (non-Javadoc)
   * @see dk.nsi.documentsharing.docsharingservices.context.DSSInvocationContextProvider#setConsentOverride(boolean)
   */
  public void setConsentOverride(boolean doOverride) {
    throw new IllegalStateException("Not supported for citizen");
  }

  /* (non-Javadoc)
   * @see dk.nsi.documentsharing.docsharingservices.context.DSSInvocationContextProvider#setFlowId(java.lang.String)
   */
  public void setFlowId(String flowId) {
    this.flowId = flowId;
  }

  /* (non-Javadoc)
   * @see dk.nsi.documentsharing.docsharingservices.context.DSSInvocationContextProvider#getSystemContextProvider()
   */
  public DSSSystemContextProvider getSystemContextProvider() {
    return systemContextProvider;
  }

  /* (non-Javadoc)
   * @see dk.nsi.documentsharing.docsharingservices.context.DSSInvocationContextProvider#getInvocationContext(dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthority)
   */
  public DSSInvocationContext getInvocationContext(ValueWithAssigningAuthority patientId) throws Exception {
    CitizenHsuid user = createUser(patientId);
    Document requestDoc = prepareRequestDocument(userContextProvider.getIDCard());

    return new DSSInvocationContext(hsuidConsumer.getAsHeader(user),
      getSecurityHeader(requestDoc),
      getMedcomHeader(requestDoc).value);
  }

  private Document prepareRequestDocument(IDCard idCard) {
    Request request = getFactory().createNewRequest(false /* do not request non-repudiation */, flowId);
    request.setIDCard(idCard);
    request.setDGWSVersion("1.0.1");
    return request.serialize2DOMDocument();
  }

  protected CitizenHsuid createUser(ValueWithAssigningAuthority patientId) {
    final String patientIdValue = (patientId != null && patientId.getValue() != null) ? patientId.getValue() : null;
    return hsuidConsumer.createCitizenUser(
      userContextProvider.getUserId(), patientIdValue, relation);
  }

  protected Security getSecurityHeader(Document doc) {
    return HeaderConverter.SECURITY.documentToJaxb(Security.class, doc);
  }

  protected Holder<Header> getMedcomHeader(Document doc) {
    return new Holder<Header>(HeaderConverter.MEDCOM.documentToJaxb(Header.class, doc));
  }

  protected SOSIFactory getFactory() {
    Properties properties = SignatureUtil.setupCryptoProviderForJVM();
    return new SOSIFactory(new EmptyCredentialVault(), properties);
  }

  @Override
  public void setParentRelation() {
    relation = LegalValues.CHILD_CUSTODY_HOLDER;
  }

  @Override
  public void setGuardianRelation() {
    relation = LegalValues.GUARDIAN;
  }

  @Override
  public void setProxyHolderRelation() {
    relation = LegalValues.PROXY_HOLDER;    
  }

  @Override
  public void setSelfRelation() {
    relation = LegalValues.CITIZEN;    
  }

}
