/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider;

import java.io.File;
import java.io.FileNotFoundException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;

import org.joda.time.DateTime;
import org.junit.Ignore;
import org.junit.Test;

import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.documentsharing.consumer.DocumentConsumer;
import dk.nsi.documentsharing.core.metadata.DocumentMetadata;
import dk.nsi.documentsharing.core.metadata.model.Person;
import dk.nsi.documentsharing.core.metadata.model.PersonImpl;
import dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthority;
import dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthorityImpl;
import dk.nsi.documentsharing.core.query.DocumentMetadataQueryResult;
import dk.nsi.documentsharing.core.query.FindDocumentsQuery;
import dk.nsi.documentsharing.core.query.FindDocumentsQueryImpl;
import dk.nsi.documentsharing.docsharingservices.ConfigurationProvider;
import dk.nsi.documentsharing.docsharingservices.DSSDocumentConsumerFactory;
import dk.nsi.documentsharing.docsharingservices.DSSUserContextProviderDGWSSample;
import dk.nsi.documentsharing.docsharingservices.FileBasedConfigurationProvider;
import dk.nsi.documentsharing.docsharingservices.context.DSSInvocationContextProvider;
import dk.nsi.documentsharing.docsharingservices.context.impl.DSSInvocationContextProviderImpl;
import dk.nsi.documentsharing.docsharingservices.context.impl.DSSSystemContextProviderImpl;
import dk.nsi.documentsharing.metadataprovider.configuration.MetadataConfiguration;
import dk.nsi.documentsharing.metadataprovider.configuration.MetadataConfigurationHandler;
import dk.nsi.documentsharing.metadataprovider.configuration.MetadataContentConfiguration;
import dk.nsi.documentsharing.util.xmlhandling.XStreamFileException;
import dk.nsi.util.Resource;
import junit.framework.Assert;

public class MetadataProviderWSIT {
  private ConfigurationProvider config = new FileBasedConfigurationProvider(System.getProperty("testclient.properties"));

  @Test
  @Ignore
  public void testSxaMetadataProviderWS() throws Exception {
    MetadataConfiguration metadataConfiguration = loadConfiguration();
    
    // Arrange
    FindDocumentsQuery query = createDefaultQuery(metadataConfiguration.getMetadataContentConfiguration());
    // Act
    DocumentMetadataQueryResult result = callWebservice(query);
    // Assert
    Assert.assertNotNull(result.getDocumentMetadataList());
    Assert.assertFalse(result.getDocumentMetadataList().isEmpty());
    Assert.assertEquals(1, result.getDocumentMetadataList().size());
    DocumentMetadata metadataResult = result.getDocumentMetadataList().get(0);
    Assert.assertNotNull(metadataResult);
    Assert.assertEquals(getPatientId(), metadataResult.getPatientId());
    if (query.getClassCodes() != null && !query.getClassCodes().isEmpty()) {
      Assert.assertEquals(query.getClassCodes().get(0), metadataResult.getClassCode());
    }
    Assert.assertEquals(query.getTypeCodes().get(0), metadataResult.getTypeCode());
    if (query.getEventCodes() != null && !query.getEventCodes().isEmpty()) {
      Assert.assertEquals(query.getEventCodes().get(0), metadataResult.getEventCodeList().get(0));
    }
    DocumentMetadataProviderTestDataHelper testHelper = new DocumentMetadataProviderTestDataHelper(new TestDataHelper());
    testHelper.compareConfigurationWithDocumentMetadata(query, metadataConfiguration, metadataResult);
  }

  @Test
  public void testMinimalQuery() throws Exception {
    // Arrange
    FindDocumentsQuery query = new FindDocumentsQueryImpl().setPatientId(getPatientId()).addQueryForOnDemandDocuments();
    // Act
    DocumentMetadataQueryResult result = callWebservice(query);
    // Assert
    Assert.assertNotNull(result.getDocumentMetadataList());
    Assert.assertTrue(result.getDocumentMetadataList().size() > 0);
    Assert.assertEquals(1, result.getDocumentMetadataList().size());
  }

  @Test
  public void startAfterStopQueryReturnsError() throws Exception {
    // Arrange
    FindDocumentsQuery query = new FindDocumentsQueryImpl().setPatientId(getPatientId()).addQueryForOnDemandDocuments();
    DateTime start = DateTime.now();
    query.setServiceStartTime(start);
    query.setServiceStopTime(start.minusDays(1)); // Stop before start
    
    // Act
    DocumentMetadataQueryResult result = callWebservice(query);
    // Assert
    Assert.assertFalse(result.hasDocumentMetadata());
    Assert.assertTrue(result.getDocumentSharingErrorList().hasErrors());
    Assert.assertTrue(result.getDocumentSharingErrorList().getErrors().stream().anyMatch(e -> e.getErrorCode().contains("ikke ligge efter")));
  }

  private MetadataConfiguration loadConfiguration()
    throws XStreamFileException, URISyntaxException, FileNotFoundException {
    File confFile = Resource.getFile(this.getClass(), "sxametadataconfiguration.xml");
    File filterFile = Resource.getFile(this.getClass(), "sxafilterconfiguration.xml");
    MetadataConfigurationHandler handler = new MetadataConfigurationHandler(confFile, filterFile);
    return handler.readConfiguration();
  }

  private ValueWithAssigningAuthority getPatientId() {
    return createDummyPerson().getId();
  }

  public Person createDummyPerson() {
    ValueWithAssigningAuthority valueAuthority = new ValueWithAssigningAuthorityImpl("2512484916", "1.2.3.4");
    return new PersonImpl(valueAuthority, "givenName", "middleName", "surName", "suffix", "prefix", "HighDegree");
  }

  public FindDocumentsQuery createDefaultQuery(MetadataContentConfiguration metadataContentConfiguration) {
    FindDocumentsQuery query = new FindDocumentsQueryImpl();
    query.addClassCode(metadataContentConfiguration.getClassCode());
    query.addTypeCode(metadataContentConfiguration.getTypeCode());
    query.addConfidentialityCode(metadataContentConfiguration.getConfidentialityCode());
    if (metadataContentConfiguration.getEventCodes() != null
      && !metadataContentConfiguration.getEventCodes().isEmpty()) {
      query.addEventCode(metadataContentConfiguration.getEventCodes().get(0));
    }
    query.addFormatCode(metadataContentConfiguration.getFormatCode());
    query.addHealtcareFacilityTypeCode(metadataContentConfiguration.getHealthcareFacilityTypeCode());
    query.addPracticeSettingCode(metadataContentConfiguration.getPracticeSettingCode());
    query.setPatientId(getPatientId());
    query.addQueryForOnDemandDocuments();
    return query;
  }

  private DocumentMetadataQueryResult callWebservice(FindDocumentsQuery query) throws Exception {
    DSSInvocationContextProvider invocationContextProvider = new DSSInvocationContextProviderImpl(
      new DSSSystemContextProviderSample(), new DSSUserContextProviderDGWSSample(config));
    DSSDocumentConsumerFactory dssDocConsumerFactory = new DSSDocumentConsumerFactory(invocationContextProvider);
    DocumentConsumer docConsumer = dssDocConsumerFactory.getInstance();
    // Perform the query
    DocumentMetadataQueryResult queryResult = docConsumer.doQuery(query);
    Assert.assertNotNull(queryResult);
    return queryResult;
  }

  protected static String getVersion() {
    String version = DSSSystemContextProviderSample.class.getPackage().getImplementationVersion();
    if (version == null || version.isEmpty()) {
      version = "no-version";
    }
    return version;
  }

  private class DSSSystemContextProviderSample extends DSSSystemContextProviderImpl {
    public DSSSystemContextProviderSample() {
      super("SystemOwner", "DSSDocumentConsumer", getVersion(), "operationsOrganisationName");
    }

    @Override
    public URL getDocumentQueryServiceEndpoint() throws MalformedURLException {
      return new URL("http://localhost:9090/sxa-documentmetadataprovider");
    }

    @Override
    public URL getDocumentRetrieveServiceEndpoint() throws MalformedURLException {
      return null;
    }

    @Override
    public URL getDocumentQueryWsdlLocation() throws MalformedURLException {
      return new URL("http://localhost:9090/sxa-documentmetadataprovider?wsdl");
    }

    @Override
    public URL getDocumentRetrieveWsdlLocation() throws MalformedURLException {
      // TODO Auto-generated method stub
      return null;
    }
  }
}
