/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider.impl;

import java.util.List;

import javax.ejb.Stateless;
import javax.inject.Inject;

import org.apache.log4j.Logger;
import org.joda.time.DateTime;

import dk.nsi.dds.projects.sxa.common.DocumentIdEncoderDecoder;
import dk.nsi.dds.projects.sxa.common.MappedCodedValue;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesMappingConfiguration;
import dk.nsi.dds.projects.sxa.common.QueryIntervalCreator;
import dk.nsi.documentsharing.core.interval.UnboundableInterval;
import dk.nsi.documentsharing.core.metadata.impl.DocumentMetadataImpl;
import dk.nsi.documentsharing.core.metadata.model.AvailabilityStatusImpl;
import dk.nsi.documentsharing.core.metadata.model.Id;
import dk.nsi.documentsharing.core.metadata.model.ObjectTypeImpl;
import dk.nsi.documentsharing.core.query.DocumentMetadataQueryResult;
import dk.nsi.documentsharing.core.query.DocumentMetadataQueryResultImpl;
import dk.nsi.documentsharing.core.query.FindDocumentsQuery;
import dk.nsi.documentsharing.id.IdConverter;
import dk.nsi.documentsharing.id.RootedId;
import dk.nsi.documentsharing.metadataprovider.DocumentMetadataProvider;
import dk.nsi.documentsharing.metadataprovider.configuration.MetadataConfiguration;
import dk.nsi.documentsharing.metadataprovider.configuration.MetadataContentConfiguration;
import dk.nsi.documentsharing.util.perflogger.PerformanceLogger;
import dk.nsi.documentsharing.util.perflogger.PerformanceLogger.Monitor;

@Stateless
public class DocumentMetadataProviderImpl implements DocumentMetadataProvider {
  private static final Logger logger = Logger.getLogger(DocumentMetadataProviderImpl.class);

  @Inject
  private MetadataConfiguration metadataConfiguration;
  @Inject
  private PracticeSettingCodesMappingConfiguration practiceSettingCodesMappingConfiguration;

  @Override
  public DocumentMetadataQueryResult doQuery(FindDocumentsQuery query) {
    try (Monitor monitor = new PerformanceLogger.Monitor("DocumentMetadataProviderImpl.doQuery")) {
      DocumentMetadataQueryResultImpl queryResult = new DocumentMetadataQueryResultImpl();
      queryResult.addEmptyList();
      try {
        if (query != null) {       
          DocumentMetadataImpl documentMetadata = getDocumentMetadata(query);
          if (documentMetadata != null) {
            queryResult.add(documentMetadata);
          }
        } else {
          logger.debug("FindDocumentsQuery is null");
          queryResult.addErrorRegistry("Forespørgslen indeholdt ikke noget data");
        }
      } catch (Exception e) {
        queryResult.addErrorRegistry(e.getMessage());
      }
      return queryResult;
    }
  }

  private DocumentMetadataImpl getDocumentMetadata(FindDocumentsQuery query) throws Exception {
    if (metadataConfiguration != null) {
      MetadataQueryFilter metadataQueryFilter = new MetadataQueryFilter(metadataConfiguration.getMetadataConfigurationFilter());
      MetadataContentConfiguration metadataContentConfiguration = metadataConfiguration.getMetadataContentConfiguration();
      
      checkIntervalAndSetDefaultValues(query);
      
      if (metadataQueryFilter != null && metadataQueryFilter.isValidQuery(query)) {
        DocumentMetadataImpl documentMetadata = new DocumentMetadataImpl();
        documentMetadata.setPatientId(query.getPatientId());
        documentMetadata.setSourcePatientId(query.getPatientId());
        documentMetadata.setDocumentUniqueId(getDocumentUniqueId(query, metadataContentConfiguration.getUniqueIdRoot()));
        documentMetadata.setAuthorPerson(metadataContentConfiguration.getAuthorPerson());
        documentMetadata.setAuthorInstitution(metadataContentConfiguration.getAuthorInstitution());
        documentMetadata.setClassCode(metadataContentConfiguration.getClassCode());
        documentMetadata.setConfidentialityCode(metadataContentConfiguration.getConfidentialityCode());
        documentMetadata.setEventCodeList(metadataContentConfiguration.getEventCodes());
        documentMetadata.setFormatCode(metadataContentConfiguration.getFormatCode());
        documentMetadata.setHealthcareFacilityTypeCode(metadataContentConfiguration.getHealthcareFacilityTypeCode());
        documentMetadata.setPracticeSettingCode(metadataContentConfiguration.getPracticeSettingCode());
        setServiceStartAndStopTime(query, documentMetadata);
        documentMetadata.setAvailabilityStatus(new AvailabilityStatusImpl(true));
        documentMetadata.setTypeCode(metadataContentConfiguration.getTypeCode());
        documentMetadata.setMimeType(metadataContentConfiguration.getMimeType());
        documentMetadata.setLanguageCode(metadataContentConfiguration.getLanguageCode());
        documentMetadata.setHomeCommunityId(metadataContentConfiguration.getHomeCommunityId());
        documentMetadata.setRepositoryUniqueId(metadataContentConfiguration.getRepositoryUniqueId());
        documentMetadata.setTitle("Lab Reports for " + (query.getPatientId() != null ? query.getPatientId().getValue() : ""));
        documentMetadata.setEntryUUID(IdConverter.convertOidToUuid(documentMetadata.getDocumentUniqueId().getId().replace(".", "")));
        documentMetadata.setObjectType(new ObjectTypeImpl(ObjectTypeImpl.ObjectTypes.OnDemand)); // Always on-demand
        return documentMetadata;
      } else {
        logger.debug("metadataQueryFilter is null or the query is invalid");
      }
    } else {
      logger.debug("MetadataConfiguration is null");
    }
    return null;
  }

  private Id getDocumentUniqueId(FindDocumentsQuery query, String uniqueIdRoot) {
    
    List<MappedCodedValue> validPracticeSettingCodes = practiceSettingCodesMappingConfiguration.getPracticeSettingCodesConfiguration().getPracticeSettingCodes();
    String encodedId = new DocumentIdEncoderDecoder(query.getPatientId().getValue(), getServiceStartDateTime(query),
      getServiceStopDateTime(query), query.getPracticeSettingCodes(), validPracticeSettingCodes).encode();
    return new RootedId(uniqueIdRoot, encodedId).getId();
  }

  private DateTime getServiceStartDateTime(FindDocumentsQuery query) {
    return getDateTime(query.getServiceStartTime(), true);
  }

  private DateTime getServiceStopDateTime(FindDocumentsQuery query) {
    return getDateTime(query.getServiceStopTime(), false);
  }

  private DateTime getDateTime(UnboundableInterval interval, boolean preferBeginning) {
    if (interval == null) {
      return null;
    }
    if (preferBeginning) {
      return (interval.getStart() != null) ? interval.getStart() : interval.getEnd();
    } else {
      return (interval.getEnd() != null) ? interval.getEnd() : interval.getStart();
    }
  }

  private void setServiceStartAndStopTime(FindDocumentsQuery query, DocumentMetadataImpl documentMetadata) {
    UnboundableInterval startStopInterval = QueryIntervalCreator.createQueryInterval(
      query.getPatientId().getValue(), getServiceStartDateTime(query), getServiceStopDateTime(query));
    documentMetadata.setServiceStartTime(startStopInterval.getStart());
    documentMetadata.setServiceStopTime(startStopInterval.getEnd());
  }
  
  private void checkIntervalAndSetDefaultValues(FindDocumentsQuery query) throws Exception {
    DateTime start = getServiceStartDateTime(query);
    DateTime end = getServiceStopDateTime(query);
    if (start != null && end != null && start.isAfter(end)) {
      throw new Exception("Tidspunktet 'Fra' (" + start.toString() + ") må ikke ligge efter 'Til' (" + end.toString() + ")");
    }
  }
}
