/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider;

import java.util.List;

import org.joda.time.DateTime;
import org.joda.time.Interval;
import org.junit.Before;
import org.junit.Test;

import dk.nsi.dds.projects.sxa.documentmetadataprovider.impl.MetadataQueryFilter;
import dk.nsi.documentsharing.core.metadata.model.CodedValue;
import dk.nsi.documentsharing.core.metadata.model.CodedValueImpl;
import dk.nsi.documentsharing.core.metadata.model.Person;
import dk.nsi.documentsharing.core.metadata.model.PersonImpl;
import dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthorityImpl;
import dk.nsi.documentsharing.core.query.FindDocumentsQuery;
import dk.nsi.documentsharing.core.query.FindDocumentsQueryImpl;
import dk.nsi.documentsharing.metadataprovider.configuration.MetadataFilterConfiguration;
import junit.framework.Assert;

public class MetadataQueryFilterTest {

  private MetadataFilterConfiguration filter;
  private MetadataFilterConfiguration valuesForQuery;
  private boolean isApproved;
  private boolean onDemandDocument;

  @Before
  public void setup() {
    filter = createDefaultFilter();
    valuesForQuery = createDefaultFilter();
    isApproved = true;
    onDemandDocument = true;
  }

  @Test
  public void queryMatchesFilter() {
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }

  @Test
  public void classCodeQueryAndFilter() {
    // Act / Assert
    valuesForQuery.getClassCodes().remove(0);
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
    filter.getClassCodes().remove(0);
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }

  @Test
  public void confidentialityCodesQueryAndFilter() {
    valuesForQuery.getConfidentialityCodes().remove(0);
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
    filter.getConfidentialityCodes().remove(0);
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }

  @Test
  public void eventCodesQueryAndFilter() {
    valuesForQuery.getEventCodes().remove(0);
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
    filter.getEventCodes().remove(0);
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }

  @Test
  public void formatCodesQueryAndFilter() {
    valuesForQuery.getFormatCodes().remove(0);
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
    filter.getFormatCodes().remove(0);
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }

  @Test
  public void typeCodesQueryAndFilter() {
    valuesForQuery.getTypeCodes().remove(0);
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
    filter.getTypeCodes().remove(0);
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }
  
  @Test
  public void practiceSettingCodesQueryAndFilter() {
    valuesForQuery.getPracticeSettingCodes().remove(0);
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
    filter.getPracticeSettingCodes().remove(0);
    valuesForQuery = createDefaultFilter();
    Assert.assertTrue(generalTest(valuesForQuery, filter, null, null, isApproved, onDemandDocument));
  }

  @Test
  public void invalidQueryTest() {
    // Arrange

    MetadataFilterConfiguration codedValuesHolder = createDefaultFilter();
    MetadataFilterConfiguration filter = createDefaultFilter();

    boolean isApproved = true;
    boolean onDemandDocument = true;
    Interval creationTime = new Interval(new DateTime(), new DateTime());
    Person authorPerson = new PersonImpl(new ValueWithAssigningAuthorityImpl("1234", "1.2.3.4"), "givenName", "middleName", "surname", "suffix", "prefix", "degree");

    // Act / Assert
    Assert.assertFalse(generalTest(codedValuesHolder, filter, authorPerson, null, isApproved, onDemandDocument));
    Assert.assertFalse(generalTest(codedValuesHolder, filter, null, creationTime, isApproved, onDemandDocument));
    Assert.assertFalse(generalTest(codedValuesHolder, filter, null, null, !isApproved, onDemandDocument));
    Assert.assertFalse(generalTest(codedValuesHolder, filter, null, null, isApproved, !onDemandDocument));

    removeAllAndAddValue(codedValuesHolder.getClassCodes(), new CodedValueImpl("asd", "asd", "asdf"));
    Assert.assertFalse(generalTest(codedValuesHolder, filter, null, null, isApproved, !onDemandDocument));

    codedValuesHolder = createDefaultFilter();
    removeAllAndAddValue(codedValuesHolder.getConfidentialityCodes(), new CodedValueImpl("asd", "asd", "asdf"));
    Assert.assertFalse(generalTest(codedValuesHolder, filter, null, null, isApproved, !onDemandDocument));

    codedValuesHolder = createDefaultFilter();
    removeAllAndAddValue(codedValuesHolder.getEventCodes(), new CodedValueImpl("asd", "asd", "asdf"));
    Assert.assertFalse(generalTest(codedValuesHolder, filter, null, null, isApproved, !onDemandDocument));

    codedValuesHolder = createDefaultFilter();
    removeAllAndAddValue(codedValuesHolder.getFormatCodes(), new CodedValueImpl("asd", "asd", "asdf"));
    Assert.assertFalse(generalTest(codedValuesHolder, filter, null, null, isApproved, !onDemandDocument));

    codedValuesHolder = createDefaultFilter();
    removeAllAndAddValue(codedValuesHolder.getTypeCodes(), new CodedValueImpl("asd", "asd", "asdf"));
    Assert.assertFalse(generalTest(codedValuesHolder, filter, null, null, isApproved, !onDemandDocument));
    
    codedValuesHolder = createDefaultFilter();
    removeAllAndAddValue(codedValuesHolder.getPracticeSettingCodes(), new CodedValueImpl("asd", "asd", "asdf"));
    Assert.assertFalse(generalTest(codedValuesHolder, filter, authorPerson, creationTime, isApproved, onDemandDocument));
  }

  private MetadataFilterConfiguration createDefaultFilter() {
    MetadataFilterConfiguration filter = new MetadataFilterConfiguration();
    filter.addClassCode(new CodedValueImpl("classCode", "classCodeSystem", "classCodeDesc"));
    filter.addTypeCode(new CodedValueImpl("typeCode", "typeCodeSystem", "typeCodeDesc"));
    filter.addEventCodes(new CodedValueImpl("eventCode", "eventCodeSystem", "eventCodeDesc"));
    filter.addConfidentialityCodes(new CodedValueImpl("confidentialityCode", "confidentialityCodeSystem", "confidentialityCodeDesc"));
    filter.addFormatCodes(new CodedValueImpl("formatCode", "formatCodeSystem", "formatCodeDesc"));
    filter.addPracticeSettingCodes(new CodedValueImpl("practiceSettingCode1", "practiceSettingCodeSystem", "practiceSettingCodeDesc"));
    filter.addPracticeSettingCodes(new CodedValueImpl("practiceSettingCode2", "practiceSettingCodeSystem", "practiceSettingCodeDesc"));
    filter.addPracticeSettingCodes(new CodedValueImpl("practiceSettingCode3", "practiceSettingCodeSystem", "practiceSettingCodeDesc"));
    return filter;
  }

  public boolean generalTest(MetadataFilterConfiguration queryCodesHolder, MetadataFilterConfiguration filter, 
    Person authorPerson, Interval creationTime, boolean isApproved, boolean onDemandDocument) {
    FindDocumentsQuery query = new FindDocumentsQueryImpl();

    for (CodedValue code : queryCodesHolder.getClassCodes()) {
      query.addClassCode(code);
    }
    for (CodedValue code : queryCodesHolder.getEventCodes()) {
      query.addEventCode(code);
    }
    for (CodedValue code : queryCodesHolder.getTypeCodes()) {
      query.addTypeCode(code);
    }
    for (CodedValue code : queryCodesHolder.getConfidentialityCodes()) {
      query.addConfidentialityCode(code);
    }
    for (CodedValue code : queryCodesHolder.getFormatCodes()) {
      query.addFormatCode(code);
    }
    for (CodedValue code : queryCodesHolder.getPracticeSettingCodes()) {
      query.addPracticeSettingCode(code);
    }
    
    if (isApproved) {
      query.addQueryForApproved();
    } else {
      query.addQueryForDeprecated();
    }
    if (onDemandDocument) {
      query.addQueryForOnDemandDocuments();
    } else {
      query.addQueryForStableDocuments();
    }

    query.setCreationTime(creationTime);
    if (authorPerson != null) {
      query.addAuthorPerson(authorPerson);
    }

    MetadataQueryFilter queryFilter = new MetadataQueryFilter(filter);

    return queryFilter.isValidQuery(query);
  }

  private void removeAllAndAddValue(List<CodedValue> list, CodedValue value) {
    list.clear();
    list.add(value);
  }
}
