/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider.impl;

import java.util.List;

import org.apache.log4j.Logger;

import dk.nsi.documentsharing.core.metadata.model.AvailabilityStatus;
import dk.nsi.documentsharing.core.metadata.model.CodedValue;
import dk.nsi.documentsharing.core.query.FindDocumentsQuery;
import dk.nsi.documentsharing.metadataprovider.configuration.MetadataFilterConfiguration;
import dk.nsi.documentsharing.metadataprovider.impl.CodedValueIgnoreDescriptionComparator;

public class MetadataQueryFilter {

  private static final Logger logger = Logger.getLogger(MetadataQueryFilter.class);
  private static final CodedValueIgnoreDescriptionComparator codedValueComparator = new CodedValueIgnoreDescriptionComparator();

  private MetadataFilterConfiguration configurationFilter;

  public MetadataQueryFilter(MetadataFilterConfiguration configurationFilter) {
    this.configurationFilter = configurationFilter;
  }

  public boolean isValidQuery(FindDocumentsQuery query) {
    if (query == null || configurationFilter == null) {
      logger.error("query is null or the filter is null");
      return false;
    }
    try {
      // Check that every condition is true
      // Throw exception otherwise
      checkTrue(query.isQueryForOnDemandDocuments(), "Not query on demand documents");
      checkTrue(checkIsApproved(query), "Query for unapproved documents");
      checkTrue(query.getCreationTime() == null || query.getCreationTime().getStart() == null, "Query has creation time start");
      checkTrue(query.getCreationTime() == null || query.getCreationTime().getEnd() == null, "Query has creation time end");
      checkTrue(query.getAuthorPersons().isEmpty(), "Query has authors");
      checkTrue(checkList(query.getClassCodes(), configurationFilter.getClassCodes()), "Query class codes not in filter");
      checkTrue(checkList(query.getTypeCodes(), configurationFilter.getTypeCodes()), "Query type codes not in filter");
      checkTrue(checkList(query.getEventCodes(), configurationFilter.getEventCodes()), "Query event codes not in filter");
      checkTrue(checkList(query.getConfidentialityCodes(), configurationFilter.getConfidentialityCodes()), "Query confidentiality codes not in filter");
      checkTrue(checkList(query.getFormatCodes(), configurationFilter.getFormatCodes()), "Query format codes not in filter");
      checkTrue(checkList(query.getPracticeSettingCodes(), configurationFilter.getPracticeSettingCodes()), "Practice Setting codes not in filter");
      return true;
    } catch (QueryInvalidException e) {
      logger.debug(String.format("Query is invalid due to the following: %s", e.getMessage()));
      return false;
    }
  }

  private boolean checkIsApproved(FindDocumentsQuery query) {
    List<AvailabilityStatus> statuses = query.getStatuses();
    for (AvailabilityStatus availabilityStatus : statuses) {
      if (availabilityStatus.isApproved()) {
        return true;
      }

    }
    return false;
  }

  private boolean checkList(List<CodedValue> queryValues, List<CodedValue> filterValues) {
    if (queryValues == null || queryValues.isEmpty() || filterValues == null || filterValues.isEmpty()) {
      return true;
    }
    if (filterValues != null) {
      for (CodedValue filterValue : filterValues) {
        for (CodedValue queryValue : queryValues) {
          if (codedValueComparator.compare(filterValue, queryValue) == 0) {
            return true;
          }
        }
      }
    }
    return false;
  }

  public void checkTrue(boolean condition, String errorMessage) throws QueryInvalidException {
    if (!condition) {
      throw new QueryInvalidException(errorMessage);
    }
  }

  private class QueryInvalidException extends RuntimeException {
    private static final long serialVersionUID = 1L;

    public QueryInvalidException(String message) {
      super(message);
    }
  }
}
