/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentmetadataprovider;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.documentsharing.core.metadata.DocumentMetadata;
import dk.nsi.documentsharing.core.metadata.model.CodedValue;
import dk.nsi.documentsharing.core.metadata.model.CodedValueImpl;
import dk.nsi.documentsharing.core.metadata.model.OrganisationImpl;
import dk.nsi.documentsharing.core.metadata.model.Person;
import dk.nsi.documentsharing.core.metadata.model.PersonImpl;
import dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthority;
import dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthorityImpl;
import dk.nsi.documentsharing.core.query.FindDocumentsQuery;
import dk.nsi.documentsharing.core.query.FindDocumentsQueryImpl;
import dk.nsi.documentsharing.metadataprovider.configuration.MetadataConfiguration;
import dk.nsi.documentsharing.metadataprovider.configuration.MetadataContentConfiguration;
import dk.nsi.documentsharing.metadataprovider.configuration.MetadataFilterConfiguration;
import junit.framework.Assert;

public class DocumentMetadataProviderTestDataHelper {

  private TestDataHelper configTestHelper;
  
  private CodedValue classCodeSxa = new CodedValueImpl("summary", "TBD", "summary");
  private CodedValue confidentialityCodeSxa = new CodedValueImpl("confidentialityCode", "confidentialityCodeSystem", "");
  public CodedValue eventCodeSxa = new CodedValueImpl("BOHA0", "TBD", "Antikoagulationsbehandling");
  public CodedValue eventCodeNonSxa = new CodedValueImpl("BOB", "TBD", "Behandling");
  private CodedValue formatCodeSxa = new CodedValueImpl("urn:ad:dk:medcom:labreports:svareksponeringsservice", "1.2.208.184.100.10", "Laboratoriesvar (samling af svar)");
  private CodedValue typeCodeSxa = new CodedValueImpl("11502-2", "2.16.840.1.113883.6.1", "LABORATORY REPORT.TOTAL");
  private CodedValue typeCodeNonSxa = new CodedValueImpl("53576-5", "2.16.840.1.113883.6.1", "Personal Health Monitoring Report");
  

  public DocumentMetadataProviderTestDataHelper(TestDataHelper configTestHelper) {
    this.configTestHelper = configTestHelper;
  }

  public FindDocumentsQuery createSxaQuery() {
    return createQuery(typeCodeSxa, null, null, null);
  }
  
  public FindDocumentsQuery createSxaQueryWithPracticeSettingCode() {
    return createQuery(typeCodeSxa, null, null, new ArrayList<>(Arrays.asList(configTestHelper.practiceSettingCodeForBiochemistry)));
  }
  
  public FindDocumentsQuery createSxaQueryWithMultiplePracticeSettingCode() {
    return createQuery(typeCodeSxa, null, null, new ArrayList<>(Arrays.asList(
      configTestHelper.practiceSettingCodeForMicrobiology,
      configTestHelper.practiceSettingCodeForBiochemistry,
      configTestHelper.practiceSettingCodeForPathology)));
  }
  
  public FindDocumentsQuery createSxaQueryWithInvalidPracticeSettingCode() {
    return createQuery(typeCodeSxa, null, null, new ArrayList<>(Arrays.asList(configTestHelper.practiceSettingCodeNonSxa)));
  }

  public FindDocumentsQuery createNonSxaQuery() {
    return createQuery(typeCodeNonSxa, null, null, null);
  }

  public FindDocumentsQuery createQueryBeyondSxaMetadata() {
    FindDocumentsQuery query = createSxaQuery();
    query.addHealtcareFacilityTypeCode(new CodedValueImpl("code6", "codesystem6", ""));
    return query;
  }

  public FindDocumentsQuery createQueryLessSpecificThanSxaMetadata() {
    FindDocumentsQuery result = createSxaQuery();
    result.getFormatCodes().clear();
    return result;
  }

  public FindDocumentsQuery createQuery(CodedValue typeCode, CodedValue classCode, CodedValue eventCode, List<CodedValue> practiceSettingCodes) {
    FindDocumentsQuery query = new FindDocumentsQueryImpl();
    query.setPatientId(createDummyPerson().getId());
    query.addQueryForOnDemandDocuments();
    if (classCode != null) {
      query.addClassCode(classCode);
    }
    query.addTypeCode(typeCode);
    query.addConfidentialityCode(confidentialityCodeSxa);
    query.addFormatCode(formatCodeSxa);
    if (eventCode != null) {
      query.addEventCode(eventCode);
    }
    
    if (practiceSettingCodes != null) {
      for(CodedValue code : practiceSettingCodes) {
        query.addPracticeSettingCode(code);
      }
    }
    
    return query;
  }

  public Person createDummyPerson() {
    ValueWithAssigningAuthority valueAuthority = new ValueWithAssigningAuthorityImpl("2512484916", "1.2.3.4");
    return new PersonImpl(valueAuthority, "givenName", "middleName", "surName", "suffix", "prefix", "HighDegree");
  }

  public MetadataConfiguration createDefaultConfiguration() {
    String expectedHomeValue = "homevalue";
    String expectedRepoValue = "repovalue";
    String expectedUniqueIdRoot = "1.2.3.4";
    MetadataContentConfiguration metadataContentConfig = new MetadataContentConfiguration();
    metadataContentConfig.setClassCode(classCodeSxa);
    metadataContentConfig.setConfidentialityCode(confidentialityCodeSxa);
    metadataContentConfig.setTypeCode(typeCodeSxa);
    metadataContentConfig.setLanguageCode("languageCode");
    metadataContentConfig.setMimeType("mimeType");
    metadataContentConfig.setPracticeSettingCode(configTestHelper.practiceSettingCodeForMicrobiology);
    metadataContentConfig.setHomeCommunityId(expectedHomeValue);
    metadataContentConfig.setHealthcareFacilityTypeCode(new CodedValueImpl("", "2.16.840.1.113883.6.96", ""));
    metadataContentConfig.setAuthorInstitution(new OrganisationImpl("institution", new ValueWithAssigningAuthorityImpl("1234", "2.16.15.16")));
    metadataContentConfig.setFormatCode(formatCodeSxa);
    metadataContentConfig.setRepositoryUniqueId(expectedRepoValue);
    metadataContentConfig.setUniqueIdRoot(expectedUniqueIdRoot);

    MetadataFilterConfiguration filterConfig = new MetadataFilterConfiguration();
    filterConfig.addClassCode(classCodeSxa);
    filterConfig.addTypeCode(typeCodeSxa);
    filterConfig.addConfidentialityCodes(confidentialityCodeSxa);
    filterConfig.addFormatCodes(formatCodeSxa);
    filterConfig.addPracticeSettingCodes(configTestHelper.practiceSettingCodeForMicrobiology);
    filterConfig.addPracticeSettingCodes(configTestHelper.practiceSettingCodeForBiochemistry);
    filterConfig.addPracticeSettingCodes(configTestHelper.practiceSettingCodeForPathology);

    MetadataConfiguration metadataConfig = new MetadataConfiguration();
    metadataConfig.setMetadataConfigurationFilter(filterConfig);
    metadataConfig.setMetadataContentConfiguration(metadataContentConfig);

    return metadataConfig;
  }

  public void assertBasicMetadata(FindDocumentsQuery query, DocumentMetadata metadata) {
    Assert.assertEquals("Lab Reports for " + query.getPatientId().getValue(), metadata.getTitle());
  }

  public void compareConfigurationWithDocumentMetadata(FindDocumentsQuery query, MetadataConfiguration sxaConfiguration, DocumentMetadata metadata) {
    MetadataContentConfiguration sxaMetadataConfiguration = sxaConfiguration.getMetadataContentConfiguration();

    if (query.getClassCodes() != null && !query.getClassCodes().isEmpty()) {
      Assert.assertTrue(compareLists(query.getClassCodes(), metadata.getClassCode()));
    }
    if (query.getEventCodes() != null && !query.getEventCodes().isEmpty())
    Assert.assertTrue(comparePartialLists(query.getEventCodes(), metadata.getEventCodeList()));
    Assert.assertTrue(compareLists(query.getTypeCodes(), metadata.getTypeCode()));
    Assert.assertTrue(compareLists(query.getStatuses(), metadata.getAvailabilityStatus()));
    Assert.assertEquals(query.getPatientId(), metadata.getPatientId());
    Assert.assertEquals(query.getPatientId(), metadata.getSourcePatientId());

    Assert.assertEquals(sxaMetadataConfiguration.getAuthorInstitution(), metadata.getAuthorInstitution());
    Assert.assertEquals(sxaMetadataConfiguration.getAuthorPerson(), metadata.getAuthorPerson()); 
    Assert.assertEquals(sxaMetadataConfiguration.getConfidentialityCode(), metadata.getConfidentialityCode());
    Assert.assertEquals(sxaMetadataConfiguration.getFormatCode(), metadata.getFormatCode());
    Assert.assertEquals(sxaMetadataConfiguration.getHealthcareFacilityTypeCode(), metadata.getHealthcareFacilityTypeCode());
    Assert.assertEquals(sxaMetadataConfiguration.getPracticeSettingCode(), metadata.getPracticeSettingCode());

    Assert.assertEquals(sxaMetadataConfiguration.getHomeCommunityId(), metadata.getHomeCommunityId());
    Assert.assertEquals(sxaMetadataConfiguration.getLanguageCode(), metadata.getLanguageCode());
    Assert.assertEquals(sxaMetadataConfiguration.getMimeType(), metadata.getMimeType());
    Assert.assertEquals(sxaMetadataConfiguration.getRepositoryUniqueId(), metadata.getRepositoryUniqueId());
    Assert.assertTrue(metadata.getDocumentUniqueId().getId().startsWith(sxaMetadataConfiguration.getUniqueIdRoot()));
  }

  private <T> boolean compareLists(List<T> expectedList, T actual) {
    List<T> actualList = new ArrayList<T>();
    actualList.add(actual);
    return compareCompleteLists(expectedList, actualList);
  }

  private <T> boolean compareCompleteLists(List<T> expectedList, List<T> actualList) {
    boolean found = false;

    if (expectedList == null || actualList == null) {
      return false;
    }
    if (expectedList.size() != actualList.size()) {
      return false;
    }
    for (T actualValue : actualList) {
      found = false;
      for (T expectedValue : expectedList) {
        if (actualValue.equals(expectedValue)) {
          found = true;
          break;
        }
      }
      if (!found) {
        return false;
      }
    }
    return found;
  }

  private <T> boolean comparePartialLists(List<T> expectedList, List<T> actualList) {
    if (expectedList == null || actualList == null) {
      return false;
    }
    if (expectedList.size() != actualList.size()) {
      return false;
    }
    for (T expectedValue : expectedList) {
      for (T actualValue : actualList) {
        if (actualValue.equals(expectedValue)) {
          return true;
        }
      }
    }
    return false;
  }

}
