/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished to do
 * so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.joda.time.DateTime;

import dk.nsi.dds.projects.sxa.common.DocumentIdEncoderDecoder;
import dk.nsi.dds.projects.sxa.common.MappedCodedValue;
import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.documentprovider.impl.ConfigurationLoader;
import dk.nsi.documentsharing.consumer.DocumentConsumer;
import dk.nsi.documentsharing.core.metadata.SimpleDocumentMetadata;
import dk.nsi.documentsharing.core.metadata.impl.SimpleDocumentMetadataImpl;
import dk.nsi.documentsharing.core.metadata.model.CodedValue;
import dk.nsi.documentsharing.core.metadata.model.Id;
import dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthority;
import dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthorityImpl;
import dk.nsi.documentsharing.core.retrieve.DocumentsToRetrieve;
import dk.nsi.documentsharing.core.retrieve.DocumentsToRetrieveImpl;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocument;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocumentsResult;
import dk.nsi.documentsharing.docsharingservices.ConfigurationProvider;
import dk.nsi.documentsharing.docsharingservices.DSSDocumentConsumerFactory;
import dk.nsi.documentsharing.docsharingservices.DSSUserContextProviderDGWSSample;
import dk.nsi.documentsharing.docsharingservices.FileBasedConfigurationProvider;
import dk.nsi.documentsharing.docsharingservices.context.DSSInvocationContextProvider;
import dk.nsi.documentsharing.docsharingservices.context.impl.DSSInvocationContextProviderImpl;
import dk.nsi.documentsharing.docsharingservices.context.impl.DSSSystemContextProviderImpl;
import dk.nsi.documentsharing.id.IdImpl;
import dk.nsi.documentsharing.id.RootedId;
import dk.nsi.documentsharing.provider.impl.ProviderConfiguration;
import dk.nsi.documentsharing.provider.impl.ProviderConfigurationHandler;
import dk.nsi.util.Resource;
import junit.framework.Assert;

public class RequestPerformer {
  private List<CodedValue> requestPracticeSettingCodes;
  private DateTime fromTime;
  private DateTime toTime;
  private String patientId;
  private String uniqueRootId;
  private TestDataHelper configTestData;
  private ProviderConfiguration providerConfiguration;
  private boolean validateResponse;
  private ConfigurationProvider config = new FileBasedConfigurationProvider(
    System.getProperty("testclient.properties"));

  public RequestPerformer() {
    this.requestPracticeSettingCodes = new ArrayList<CodedValue>();
    this.validateResponse = false;
    this.uniqueRootId = "1.2.3.4.55.66";
    this.configTestData = new TestDataHelper();
    this.providerConfiguration = loadConfiguration();
  }

  public RequestPerformer addRequestPracticeSettingsCodes(CodedValue... requestPracticeSettingsCodes) {
    Arrays.stream(requestPracticeSettingsCodes).forEach(code -> this.requestPracticeSettingCodes.add(code));
    return this;
  }

  public RequestPerformer providerConfiguration(ProviderConfiguration providerConfiguration) {
    this.providerConfiguration = providerConfiguration;
    return this;
  }

  public RequestPerformer from(DateTime fromTime) {
    this.fromTime = fromTime;
    return this;
  }

  public RequestPerformer to(DateTime toTime) {
    this.toTime = toTime;
    return this;
  }

  public RequestPerformer patientId(String patientId) {
    this.patientId = patientId;
    return this;
  }

  public RequestPerformer rootId(String uniqueRootId) {
    this.uniqueRootId = uniqueRootId;
    return this;
  }

  public ResponseContext performRequest() throws Exception {
    List<MappedCodedValue> validPracticeSettingCodes = configTestData.getMappedPracticeSettingCodes();
    Id docid = createDocumentUniqueId(patientId, fromTime, toTime, uniqueRootId, requestPracticeSettingCodes,
      validPracticeSettingCodes);
    ValueWithAssigningAuthority patientIdValue = new ValueWithAssigningAuthorityImpl(patientId,
      new IdImpl("1.208.3.4").getId());
    DocumentsToRetrieve request = createDefaultRequest(providerConfiguration, patientIdValue, docid);
    return validateResponse ? callWebserviceAndValidateResponse(docid, request) : callWebservice(docid, request);
  }

  private Id createDocumentUniqueId(
    String patientId, DateTime from, DateTime to, String uniqueIdRoot, List<CodedValue> practiceSettingCodes,
    List<MappedCodedValue> validPracticeSettingCodes) {
    String encodedId = new DocumentIdEncoderDecoder(patientId, from, to, practiceSettingCodes,
      validPracticeSettingCodes).encode();
    return new RootedId(uniqueIdRoot, encodedId).getId();
  }

  private ProviderConfiguration loadConfiguration() {
    try {
      File confFile = Resource.getFile(this.getClass(), ConfigurationLoader.PROVIDER_CONFIGURATION_FILE_NAME);
      ProviderConfigurationHandler handler = new ProviderConfigurationHandler(confFile);
      return handler.readConfiguration();
    } catch (Exception ex) {
      throw new RuntimeException(ex);
    }
  }

  public DocumentsToRetrieve createDefaultRequest(
    ProviderConfiguration providerConfiguration, ValueWithAssigningAuthority patientId, Id uniqueDocumentId) {
    List<SimpleDocumentMetadata> metadataList = new ArrayList<SimpleDocumentMetadata>();
    metadataList.add(createMetadataDocument(patientId, providerConfiguration.getRepositoryUniqueId(),
      providerConfiguration.getHomeCommunityId(), uniqueDocumentId));
    return new DocumentsToRetrieveImpl(metadataList);
  }

  private SimpleDocumentMetadata createMetadataDocument(
    ValueWithAssigningAuthority patientId, String repositoryUniqueId, String homeCommunityId, Id documentUniqueId) {
    SimpleDocumentMetadataImpl metadata = new SimpleDocumentMetadataImpl();
    metadata.setDocumentUniqueId(documentUniqueId);
    metadata.setHomeCommunityId(homeCommunityId);
    metadata.setPatientId(patientId);
    metadata.setRepositoryUniqueId(repositoryUniqueId);
    return metadata;
  }

  private ResponseContext callWebservice(Id docid, DocumentsToRetrieve documentsToRetrieve) throws Exception {
    DSSInvocationContextProvider invocationContextProvider = new DSSInvocationContextProviderImpl(
      new DSSSystemContextProviderSample(), new DSSUserContextProviderDGWSSample(config));
    DSSDocumentConsumerFactory dssDocConsumerFactory = new ProviderDSSDocumentConsumerFactory(
      invocationContextProvider);
    DocumentConsumer docConsumer = dssDocConsumerFactory.getInstance();
    // Perform the query
    ResponseContext context = new ResponseContext();
    context.setGeneratedDocumentId(docid);
    context.setHomeCommunity(providerConfiguration.getHomeCommunityId());
    context.setRepositoryId(providerConfiguration.getRepositoryUniqueId());
    context.setResult(docConsumer.doRetrieve(documentsToRetrieve));
    return context;
  }
  
  private ResponseContext callWebserviceAndValidateResponse(Id docid, DocumentsToRetrieve request) throws Exception {
    ResponseContext context = callWebservice(docid, request);
    RetrievedDocumentsResult result = context.getResult();
    
    Assert.assertNotNull(result);
    Assert.assertNotNull(result.getDocuments());
    Assert.assertEquals(1, result.getDocuments().size());
    Assert.assertNotNull(result.getDocumentSharingErrorList());
    Assert.assertTrue(result.getDocumentSharingErrorList().getAll().isEmpty());
    Assert.assertTrue(result.getUnreachableDocuments().isEmpty());
    RetrievedDocument retrievedDocument = result.getDocuments().get(0);
    Assert.assertNotNull(retrievedDocument.getDocument());
    Assert.assertEquals(context.getGeneratedDocumentId(), retrievedDocument.getDocumentUniqueId());
    Assert.assertEquals(context.getHomeCommunity(), retrievedDocument.getHomeCommunityId());
    Assert.assertEquals("text/xml", retrievedDocument.getMimeType());
    Assert.assertEquals(context.getRepositoryId(), retrievedDocument.getRepositoryUniqueId());
    return context;
  }

  protected static String getVersion() {
    String version = DSSSystemContextProviderSample.class.getPackage().getImplementationVersion();
    if (version == null || version.isEmpty()) {
      version = "no-version";
    }
    return version;
  }

  private class DSSSystemContextProviderSample extends DSSSystemContextProviderImpl {
    public DSSSystemContextProviderSample() {
      super("SystemOwner", "DSSDocumentConsumer", getVersion(), "operationsOrganisationName");
    }

    @Override
    public URL getDocumentQueryServiceEndpoint() throws MalformedURLException {
      return null;
    }

    @Override
    public URL getDocumentRetrieveServiceEndpoint() throws MalformedURLException {
      return new URL("http://localhost:9090/sxa-documentprovider");
    }

    @Override
    public URL getDocumentQueryWsdlLocation() throws MalformedURLException {
      return new URL("http://localhost:9090/sxa-documentmetadataprovider?wsdl");
    }

    @Override
    public URL getDocumentRetrieveWsdlLocation() throws MalformedURLException {
      return new URL("http://localhost:9090/sxa-documentprovider?wsdl");
    }
  }

  public RequestPerformer expectOnePIWDocument() {
    this.validateResponse = true;
    return this;
  }
}
