/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.joda.time.DateTime;
import org.junit.Ignore;
import org.junit.Test;

import com.cgi.svareksponering.ResultatPakkePIH;

import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.documentprovider.impl.ConfigurationLoader;
import dk.nsi.dds.projects.sxa.labreportservicestub.TestData;
import dk.nsi.documentsharing.core.error.DocumentSharingError;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocumentsResult;
import dk.nsi.documentsharing.provider.impl.ProviderConfiguration;
import dk.nsi.documentsharing.provider.impl.ProviderConfigurationHandler;
import dk.nsi.documentsharing.util.xmlhandling.XStreamFileException;
import dk.nsi.schema.privacy.pih._1.PrivacyInformationHeaderType;
import dk.nsi.schema.privacy.piw._1.PrivacyInformationWrapperType;
import dk.nsi.util.Resource;
import junit.framework.Assert;

/**
 * 
 * These tests works for both CGI test service and provided stub
 * 
 * Note that the CGI test service returns the same data response despite choice of parameters
 *
 */
public class DocumentProviderWSIT {
  private JAXBContext jaxbContext = createJAXBContext();
  private TestDataHelper configTestData = new TestDataHelper();

  @Test
  public void requestEmptySpeciale() throws Exception {
    // No practiceSettingsCodes added
    // Arrange
    ResponseContext context = new RequestPerformer().patientId(TestData.DEFAULT_PATIENT_ID)
      .from(new DateTime())
      .to(new DateTime().plusMinutes(1))
      .performRequest();
    // Assert
    Assert.assertNotNull(unmarshalPIW(context.getResult().getDocuments().get(0).getDocument()));
  }

  @Test
  public void requestSuccessSingleSpeciale() throws Exception {
    // Arrange
    ResponseContext context = new RequestPerformer().patientId(TestData.DEFAULT_PATIENT_ID)
      .from(new DateTime())
      .to(new DateTime().plusMinutes(1))
      .addRequestPracticeSettingsCodes(configTestData.practiceSettingCodeForMicrobiology)
      .expectOnePIWDocument()
      .performRequest();
    // Assert
    Assert.assertNotNull(unmarshalPIW(context.getResult().getDocuments().get(0).getDocument()));
  }

  @Test
  @Ignore
  public void requestSuccessTwoSpeciale() throws Exception {
    // Arrange
    ResponseContext context = new RequestPerformer().patientId(TestData.DEFAULT_PATIENT_ID)
      .from(new DateTime())
      .to(new DateTime().plusMinutes(1))
      .addRequestPracticeSettingsCodes(configTestData.practiceSettingCodeForMicrobiology,
        configTestData.practiceSettingCodeForBiochemistry)
      .expectOnePIWDocument()
      .performRequest();
    // Assert
    PrivacyInformationWrapperType piw = unmarshalPIW(context.getResult().getDocuments().get(0).getDocument());
    Assert.assertNotNull(piw);
    expectedDocumentCountInPIW(piw, 3); // CGI test always returns same data....
  }

  private void expectedDocumentCountInPIW(PrivacyInformationWrapperType piw, int expectedCount) {
    Assert.assertEquals(expectedCount, piw.getHeader().getDeclaration().size());
  }

  @Test
  public void requestFromTimeIsNull() throws Exception {
    // Arrange
    ResponseContext context = new RequestPerformer().patientId(TestData.DEFAULT_PATIENT_ID)
      .from(null)
      .to(new DateTime().plusMinutes(1))
      .addRequestPracticeSettingsCodes(configTestData.practiceSettingCodeForMicrobiology)
      .performRequest();
    // Assert
    Assert.assertNotNull(context.getResult().getDocuments());
  }

  @Test
  public void requestToTimeIsNull() throws Exception {
    // Arrange
    ResponseContext context = new RequestPerformer().patientId(TestData.DEFAULT_PATIENT_ID)
      .from(new DateTime())
      .to(null)
      .addRequestPracticeSettingsCodes(configTestData.practiceSettingCodeForMicrobiology)
      .performRequest();
    // Assert
    Assert.assertNotNull(context.getResult().getDocuments());
  }
  
  @Test
  public void requestFromAndToTimesAreNull() throws Exception {
    // Arrange
    ResponseContext context = new RequestPerformer().patientId(TestData.DEFAULT_PATIENT_ID)
      .from(null)
      .to(null)
      .addRequestPracticeSettingsCodes(configTestData.practiceSettingCodeForMicrobiology)
      .performRequest();
    // Assert
    Assert.assertNotNull(context.getResult().getDocuments());
  }

  @Test
  public void requestFromTimeAfterToTime() throws Exception {
    // Arrange
    ResponseContext context = new RequestPerformer().patientId(TestData.DEFAULT_PATIENT_ID)
      .from(new DateTime().plusMinutes(1))
      .to(new DateTime())
      .addRequestPracticeSettingsCodes(configTestData.practiceSettingCodeForMicrobiology)
      .performRequest();
    // Assert
    expectRepositoryError(context);
  }

  @Test
  public void requestInvalidPatientId() throws Exception {
    // Arrange
    String invalidPatientIdWithTooManyDigits = "25124849161111111";
    ResponseContext context = new RequestPerformer().patientId(invalidPatientIdWithTooManyDigits)
      .from(new DateTime())
      .to(new DateTime().plusMinutes(1))
      .addRequestPracticeSettingsCodes(configTestData.practiceSettingCodeForMicrobiology)
      .performRequest();
    // Assert
    expectRepositoryError(context);
  }

  @Test
  public void requestWrongHomeCommunityId() throws Exception {
    // Arrange
    // Override home community for the request
    ProviderConfiguration providerConfiguration = loadConfiguration();
    String wrongHomeCommunityId = providerConfiguration.getHomeCommunityId() + ".3.1";
    providerConfiguration.setHomeCommunityId(wrongHomeCommunityId);
    // Act
    ResponseContext context = new RequestPerformer().patientId(TestData.DEFAULT_PATIENT_ID)
      .from(null)
      .to(null)
      .providerConfiguration(providerConfiguration)
      .performRequest();
    // Assert
    expectDocumentSharingErrorWithWrongCommunityId (wrongHomeCommunityId, context);
  }
  
  private void expectRepositoryError(ResponseContext context) {
    RetrievedDocumentsResult result = context.getResult();
    Assert.assertNotNull(result);
    Assert.assertNull(result.getDocuments());
    Assert.assertNotNull(result.getDocumentSharingErrorList());
    Assert.assertNotNull(result.getDocumentSharingErrorList().getAll());
    Assert.assertFalse(result.getDocumentSharingErrorList().getAll().isEmpty());
    Assert.assertEquals(1, result.getDocumentSharingErrorList().getAll().size());
    DocumentSharingError error = result.getDocumentSharingErrorList().getAll().get(0);
    Assert.assertEquals(context.getGeneratedDocumentId().toString(), error.getDocumentUniqueIdFromContext());
    Assert.assertEquals(context.getRepositoryId(), error.getRepositoryUniqueIdFromContext());
    Assert.assertTrue(error.isSeverityError());
    Assert.assertTrue(error.isStandardError());
    Assert.assertEquals(DocumentSharingError.ErrorCode.RepositoryError, error.getStandardErrorCode());
  }

  private void expectDocumentSharingErrorWithWrongCommunityId (
    String expectedHomeCommunityId, ResponseContext context) {
    RetrievedDocumentsResult result = context.getResult();
    Assert.assertNotNull(result);
    Assert.assertNull(result.getDocuments());
    Assert.assertNotNull(result.getDocumentSharingErrorList());
    Assert.assertNotNull(result.getDocumentSharingErrorList().getAll());
    Assert.assertFalse(result.getDocumentSharingErrorList().getAll().isEmpty());
    Assert.assertEquals(2, result.getDocumentSharingErrorList().getAll().size());
    DocumentSharingError error = result.getDocumentSharingErrorList().getAll().get(1);
    Assert.assertEquals(expectedHomeCommunityId, error.getHomeCommunityIdFromContext());
    Assert.assertEquals(context.getGeneratedDocumentId().toString(), error.getDocumentUniqueIdFromContext());
    Assert.assertEquals(context.getRepositoryId(), error.getRepositoryUniqueIdFromContext());
    Assert.assertTrue(error.isSeverityError());
    Assert.assertTrue(error.isStandardError());
    Assert.assertEquals(DocumentSharingError.ErrorCode.UnknownCommunity, error.getStandardErrorCode());
  }

  private ProviderConfiguration loadConfiguration() throws XStreamFileException, FileNotFoundException {
    File confFile = Resource.getFile(this.getClass(), ConfigurationLoader.PROVIDER_CONFIGURATION_FILE_NAME);
    ProviderConfigurationHandler handler = new ProviderConfigurationHandler(confFile);
    return handler.readConfiguration();
  }

  private PrivacyInformationWrapperType unmarshalPIW(byte[] data) throws JAXBException {
    Source source = new StreamSource(new InputStreamReader(new ByteArrayInputStream(data), StandardCharsets.UTF_8));
    JAXBElement<PrivacyInformationWrapperType> piwElement = jaxbContext.createUnmarshaller().unmarshal(source,
      PrivacyInformationWrapperType.class);
    return piwElement.getValue();
  }

  private JAXBContext createJAXBContext() {
    try {
      return JAXBContext.newInstance(PrivacyInformationWrapperType.class, PrivacyInformationHeaderType.class,
        ResultatPakkePIH.class);
    } catch (JAXBException e) {
      throw new RuntimeException(e);
    }
  }
}
