/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.enterprise.inject.Produces;
import javax.inject.Inject;
import javax.xml.ws.Holder;

import org.jglue.cdiunit.CdiRunner;
import org.joda.time.DateTime;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi.common.util.SLALogItemFactory;
import dk.nsi.dds.projects.sxa.common.DocumentIdEncoderDecoder;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesConfiguration;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesMappingConfiguration;
import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallException;
import dk.nsi.dds.projects.sxa.documentprovider.impl.SXADocumentProviderConfig.SXAConfigKeys;
import dk.nsi.dds.projects.sxa.documentprovider.invoker.SXAInvokeHandler;
import dk.nsi.dds.projects.sxa.documentprovider.util.SXAInvocationAttributes;
import dk.nsi.dds.projects.sxa.documentprovider.util.TestConfig;
import dk.nsi.documentsharing.core.error.DocumentSharingError;
import dk.nsi.documentsharing.core.error.DocumentSharingError.ErrorCode;
import dk.nsi.documentsharing.core.metadata.SimpleDocumentMetadata;
import dk.nsi.documentsharing.core.metadata.impl.SimpleDocumentMetadataImpl;
import dk.nsi.documentsharing.core.metadata.model.CodedValue;
import dk.nsi.documentsharing.core.metadata.model.CodedValueImpl;
import dk.nsi.documentsharing.core.metadata.model.Id;
import dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthority;
import dk.nsi.documentsharing.core.metadata.model.ValueWithAssigningAuthorityImpl;
import dk.nsi.documentsharing.core.retrieve.DocumentsToRetrieveImpl;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocument;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocumentsResult;
import dk.nsi.documentsharing.docsharingservices.context.DSSInvocationContext;
import dk.nsi.documentsharing.id.IdImpl;
import dk.nsi.documentsharing.id.RootedId;
import dk.nsi.documentsharing.provider.impl.ProviderConfiguration;
import dk.nsi.documentsharing.provider.impl.ProviderConfigurationHandler;
import dk.nsi.documentsharing.util.xmlhandling.XStreamFileException;
import dk.nsi.hsuid.ActingUserCivilRegistrationNumberAttribute;
import dk.nsi.hsuid.Attribute;
import dk.nsi.hsuid.CitizenCivilRegistrationNumberAttribute;
import dk.nsi.hsuid.HealthcareServiceUserIdentificationHeaderUtil;
import dk.nsi.hsuid.OperationsOrganisationNameAttribute;
import dk.nsi.hsuid.SystemNameAttribute;
import dk.nsi.hsuid.SystemVendorNameAttribute;
import dk.nsi.hsuid.SystemVersionAttribute;
import dk.nsi.hsuid.UserTypeAttribute;
import dk.nsi.hsuid._2016._08.hsuid_1_1.HsuidHeader;
import dk.nsi.util.Config;
import junit.framework.Assert;

@RunWith(CdiRunner.class)
public class DocumentRetrieverImplTest {
  @Produces
  @Mock
  private SXAInvokeHandler sxaInvokeHandler;
  @Produces
  @Mock
  private ProviderConfigurationHandler providerConfigurationHandler;
  @Produces
  @Mock
  private PracticeSettingCodesMappingConfiguration practiceSettingCodesMappingConfiguration;
  @Inject
  private DocumentRetrieverImpl documentRetriever;
  @Mock
  private DSSInvocationContext invocationContext;
  @Produces
  @Mock
  private SLALogItemFactory slaLogItemFactory;
  
  // Decoded patient id (from DocumentUniqueId) and patient id simulated from HSUID header
  // (patient id in SimpleDocumentMetadata) differ
  @Test
  public void retrieveDocuments_PatientIdDifferingFails() throws Exception {
    // Arrange
    String patientIdText = "1234";
    String encodedPatientId = "789";
    ValueWithAssigningAuthority patientId = 
        new ValueWithAssigningAuthorityImpl(patientIdText, new IdImpl("patientIdAssigningAuthority").getId());
    List<Id> ids = createDocumentUniqueIds(1, "1.2.3", encodedPatientId, new ArrayList<>());
    List<SimpleDocumentMetadata> simpleDocumentMetadataList = 
        createDocumentMetadataList("home", "repo", ids, patientId);
    DocumentsToRetrieveImpl documentsToRetrieve = new DocumentsToRetrieveImpl();
    documentsToRetrieve.add(simpleDocumentMetadataList);
    // Act
    RetrievedDocumentsResult result = documentRetriever.retrieveDocuments(documentsToRetrieve, createTestInvocationContext());
    // Assert
    Assert.assertNotNull(result);
    Assert.assertTrue(result.isStatusFailed());
    Assert.assertNull(result.getDocuments());
    Assert.assertTrue(result.getUnreachableDocuments().isEmpty());
    Assert.assertTrue(result.getDocumentSharingErrorList().getAll().stream()
      .anyMatch(e -> e.getStandardErrorCode().equals(ErrorCode.RepositoryError)));
  }

  @Test
  public void retrieveDocuments_OnlyReachableDocumentsTest() throws Exception {
    // Arrange
    final int documentsCount = 5;
    String homeCommunityId = "home";
    String repositoryUniqueId = "repo";
    String uniqueIdRoot = "1.2.3";
    Id id = new IdImpl("1.2.3.4");
    String patientIdText = "1234";
    byte[] data = new byte[0];
    ValueWithAssigningAuthority patientId = new ValueWithAssigningAuthorityImpl(patientIdText, id.getId());
    setupMocks(homeCommunityId, repositoryUniqueId, uniqueIdRoot);
    // Arrange Mocks
    List<Id> ids = createDocumentUniqueIds(documentsCount, uniqueIdRoot, patientIdText, new ArrayList<>());
    List<SimpleDocumentMetadata> simpleDocumentMetadataList = createDocumentMetadataList(homeCommunityId,
      repositoryUniqueId, ids, patientId);
    setupLabInvokerMock(simpleDocumentMetadataList, new ArrayList<SimpleDocumentMetadata>(), data);
    DocumentsToRetrieveImpl documentsToRetrieve = new DocumentsToRetrieveImpl();
    documentsToRetrieve.add(simpleDocumentMetadataList);
    DSSInvocationContext invocationContext = createTestInvocationContext();
    // Act
    RetrievedDocumentsResult result = documentRetriever.retrieveDocuments(documentsToRetrieve, invocationContext);
    // Assert
    Assert.assertNotNull(result);
    Assert.assertNotNull(result.getDocuments());
    Assert.assertEquals(simpleDocumentMetadataList.size(), result.getDocuments().size());
    Assert.assertNotNull(result.getDocumentSharingErrorList());
    Assert.assertTrue(result.getDocumentSharingErrorList().getAll().isEmpty());
    Assert.assertTrue(result.getUnreachableDocuments().isEmpty());
    assertReachableDocuments(result, simpleDocumentMetadataList.size(),
      providerConfigurationHandler.readConfiguration(), ids, data);
  }

  private void setupMocks(String homeCommunityId, String repositoryUniqueId, String uniqueIdRoot)
    throws XStreamFileException {
    ProviderConfiguration configuration = new ProviderConfiguration(homeCommunityId, repositoryUniqueId, uniqueIdRoot);
    Mockito.reset(providerConfigurationHandler);
    Mockito.when(providerConfigurationHandler.readConfiguration()).thenReturn(configuration);
    PracticeSettingCodesConfiguration practiceSettingCodesConfiguration = new PracticeSettingCodesConfiguration();
    practiceSettingCodesConfiguration.setPracticeSettingCodes(new TestDataHelper().getMappedPracticeSettingCodes());
    Mockito.when(practiceSettingCodesMappingConfiguration.getPracticeSettingCodesConfiguration())
      .thenReturn(practiceSettingCodesConfiguration);
  }

  private List<Id> createDocumentUniqueIds(
    int idCount, String uniqueIdRoot, String patientId, List<CodedValue> selectedPracticeCodes) {
    List<Id> ids = new ArrayList<Id>();
    DateTime now = new DateTime();
    for (int i = 0; i < idCount; i++) {
      ids.add(
        new RootedId(uniqueIdRoot, new DocumentIdEncoderDecoder(patientId, now.plusMinutes(i), now.plusMinutes(i + 1),
          selectedPracticeCodes, new TestDataHelper().getMappedPracticeSettingCodes()).encode()).getId());
    }
    return ids;
  }

  private List<Id> createDocumentUniqueIdsWithPracticeSettingCodes(int idCount, String uniqueIdRoot, String patientId) {
    List<Id> ids = new ArrayList<Id>();
    DateTime now = new DateTime();
    CodedValue psc = new CodedValueImpl("408454008", "SOR Test", "Test");
    for (int i = 0; i < idCount; i++) {
      ids.add(new RootedId(uniqueIdRoot,
        new DocumentIdEncoderDecoder(patientId, now.plusMinutes(i), null, Arrays.asList(psc),
          practiceSettingCodesMappingConfiguration.getPracticeSettingCodesConfiguration().getPracticeSettingCodes())
            .encode()).getId());
    }
    return ids;
  }

  private List<SimpleDocumentMetadata> createDocumentMetadataList(
    String homeCommunityId, String repositoryUniqueId, List<Id> ids, ValueWithAssigningAuthority patientId) {
    List<SimpleDocumentMetadata> simpleDocumentMetadataList = new ArrayList<SimpleDocumentMetadata>();
    for (int i = 0; i < ids.size(); i++) {
      simpleDocumentMetadataList
        .add(createMetadataDocument(patientId, repositoryUniqueId, homeCommunityId, ids.get(i)));
    }
    return simpleDocumentMetadataList;
  }

  private SimpleDocumentMetadata createMetadataDocument(
    ValueWithAssigningAuthority patientId, String repositoryUniqueId, String homeCommunityId, Id documentUniqueId) {
    SimpleDocumentMetadataImpl metadata = new SimpleDocumentMetadataImpl();
    metadata.setDocumentUniqueId(documentUniqueId);
    metadata.setHomeCommunityId(homeCommunityId);
    metadata.setPatientId(patientId);
    metadata.setRepositoryUniqueId(repositoryUniqueId);
    return metadata;
  }

  private void setupLabInvokerMock(
    List<SimpleDocumentMetadata> simpleDocumentMetadataList, List<SimpleDocumentMetadata> unreachableList,
    byte[] documentData) throws Exception {
    for (SimpleDocumentMetadata simpleDocumentMetadata : simpleDocumentMetadataList) {
      String documentId = simpleDocumentMetadata.getDocumentUniqueId().getId();
      Mockito.when(sxaInvokeHandler.getDocument(Mockito.eq(documentId), Mockito.any(SXAInvocationAttributes.class), Mockito.any(DSSInvocationContext.class)))
        .thenReturn(documentData);
    }
    for (SimpleDocumentMetadata simpleDocumentMetadata : unreachableList) {
      String documentId = simpleDocumentMetadata.getDocumentUniqueId().getId();
      Mockito.when(sxaInvokeHandler.getDocument(Mockito.eq(documentId), Mockito.any(SXAInvocationAttributes.class), Mockito.any(DSSInvocationContext.class)))
        .thenThrow(new InvokerCallException(new Exception()));
    }
  }

  private void assertReachableDocuments(
    RetrievedDocumentsResult retrievedDocumentsResult, int expectedReachableDocumentsCount,
    ProviderConfiguration providerConfiguration, List<Id> ids, byte[] documentData) {
    Assert.assertNotNull(retrievedDocumentsResult);
    Assert.assertNotNull(retrievedDocumentsResult.getDocuments());
    Assert.assertEquals(expectedReachableDocumentsCount, retrievedDocumentsResult.getDocuments().size());
    for (int i = 0; i < retrievedDocumentsResult.getDocuments().size(); i++) {
      RetrievedDocument retrievedDocument = retrievedDocumentsResult.getDocuments().get(i);
      Assert.assertEquals(new String(documentData), new String(retrievedDocument.getDocument()));
      Assert.assertTrue(ids.contains(retrievedDocument.getDocumentUniqueId()));
      Assert.assertEquals(providerConfiguration.getHomeCommunityId(), retrievedDocument.getHomeCommunityId());
      Assert.assertEquals("text/xml", retrievedDocument.getMimeType());
      Assert.assertEquals(providerConfiguration.getRepositoryUniqueId(), retrievedDocument.getRepositoryUniqueId());
    }
  }

  @Test
  public void retrieveDocuments_OnlyErrorsForDocumentsTest() throws Exception {
    // Arrange
    final int documentsCount = 5;
    String homeCommunityId = "home";
    String repositoryUniqueId = "repo";
    String uniqueIdRoot = "1.2.3";
    Id id = new IdImpl("1.2.3.4");
    String patientIdText = "1234";
    byte[] data = new byte[0];
    ValueWithAssigningAuthority patientId = new ValueWithAssigningAuthorityImpl(patientIdText, id.getId());
    setupMocks(homeCommunityId, repositoryUniqueId, uniqueIdRoot);
    // Arrange Mocks
    List<Id> ids = createDocumentUniqueIds(documentsCount, uniqueIdRoot, patientIdText, new ArrayList<>());
    List<SimpleDocumentMetadata> simpleDocumentMetadataList = createDocumentMetadataList(homeCommunityId,
      repositoryUniqueId, ids, patientId);
    setupLabInvokerMock(new ArrayList<SimpleDocumentMetadata>(), simpleDocumentMetadataList, data);
    DocumentsToRetrieveImpl documentsToRetrieve = new DocumentsToRetrieveImpl();
    documentsToRetrieve.add(simpleDocumentMetadataList);
    DSSInvocationContext invocationContext = createTestInvocationContext();
    // Act
    RetrievedDocumentsResult result = documentRetriever.retrieveDocuments(documentsToRetrieve, invocationContext);
    // Assert
    Assert.assertNotNull(result);
    Assert.assertNull(result.getDocuments());
    Assert.assertNotNull(result.getUnreachableDocuments());
    Assert.assertNotNull(result.getDocumentSharingErrorList());
    Assert.assertEquals(simpleDocumentMetadataList.size(), result.getDocumentSharingErrorList().getAll().size());
    ProviderConfiguration providerConfiguration = providerConfigurationHandler.readConfiguration();
    for (int i = 0; i < result.getDocumentSharingErrorList().getAll().size(); i++) {
      DocumentSharingError error = result.getDocumentSharingErrorList().getAll().get(i);
      assertDocumentIdInList(ids, error.getDocumentUniqueIdFromContext());
      Assert.assertEquals(providerConfiguration.getHomeCommunityId(), error.getHomeCommunityIdFromContext());
      Assert.assertEquals(providerConfiguration.getRepositoryUniqueId(), error.getRepositoryUniqueIdFromContext());
    }
  }

  private void assertDocumentIdInList(List<Id> ids, String documentUniqueIdFromContext) {
    Assert.assertTrue(ids.stream().anyMatch(id -> id.getId().compareTo(documentUniqueIdFromContext) == 0));
    
  }

  private void assertErrorsForDocuments(
    RetrievedDocumentsResult retrievedDocumentsResult, int expectedUnreachableDocumentsCount,
    ProviderConfiguration providerConfiguration, List<Id> ids) {
    Assert.assertNotNull(retrievedDocumentsResult);
    Assert.assertNotNull(retrievedDocumentsResult.getDocumentSharingErrorList());
    Assert.assertEquals(expectedUnreachableDocumentsCount, retrievedDocumentsResult.getDocumentSharingErrorList().getAll().size());
    for (int i = 0; i < retrievedDocumentsResult.getDocumentSharingErrorList().getAll().size(); i++) {
      DocumentSharingError error = retrievedDocumentsResult.getDocumentSharingErrorList().getAll().get(i);
      assertDocumentIdInList(ids, error.getDocumentUniqueIdFromContext());
      Assert.assertEquals(providerConfiguration.getHomeCommunityId(), error.getHomeCommunityIdFromContext());
      Assert.assertEquals(providerConfiguration.getRepositoryUniqueId(), error.getRepositoryUniqueIdFromContext());
    }
  }

  @Test
  public void retrieveDocuments_BothDocumentsTest() throws Exception {
    // Arrange
    final int documentsCount = 5;
    String homeCommunityId = "home";
    String repositoryUniqueId = "repo";
    String uniqueIdRoot = "1.2.3";
    Id id = new IdImpl("1.2.3.4");
    String patientIdText = "1234";
    ValueWithAssigningAuthority patientId = new ValueWithAssigningAuthorityImpl(patientIdText, id.getId());
    byte[] data = new byte[0];
    // Arrange Mocks
    // Half of ids are used for reachable.
    // Other half unreachable
    setupMocks(homeCommunityId, repositoryUniqueId, uniqueIdRoot);
    List<Id> ids = createDocumentUniqueIds(documentsCount * 2, uniqueIdRoot, patientIdText, new ArrayList<>());
    List<SimpleDocumentMetadata> reachableSimpleDocumentMetadataList = createDocumentMetadataList(homeCommunityId,
      repositoryUniqueId, ids.subList(0, ids.size() / 2), patientId);
    List<SimpleDocumentMetadata> unreachableSimpleDocumentMetadataList = createDocumentMetadataList(homeCommunityId,
      repositoryUniqueId, ids.subList(ids.size() / 2, ids.size()), patientId);
    setupLabInvokerMock(unreachableSimpleDocumentMetadataList, reachableSimpleDocumentMetadataList, data);
    DocumentsToRetrieveImpl documentsToRetrieve = new DocumentsToRetrieveImpl();
    documentsToRetrieve.add(reachableSimpleDocumentMetadataList);
    documentsToRetrieve.add(unreachableSimpleDocumentMetadataList);
    DSSInvocationContext invocationContext = createTestInvocationContext();
    // Act
    RetrievedDocumentsResult result = documentRetriever.retrieveDocuments(documentsToRetrieve, invocationContext);
    // Assert
    Assert.assertNotNull(result);
    Assert.assertNotNull(result.getDocuments());
    ProviderConfiguration providerConfiguration = providerConfigurationHandler.readConfiguration();
    assertErrorsForDocuments(result, reachableSimpleDocumentMetadataList.size(), providerConfiguration, ids);
    assertReachableDocuments(result, unreachableSimpleDocumentMetadataList.size(), providerConfiguration, ids, data);
  }

  /**
   * @param patientId
   * @return dummy DSSInvocationContext for testing
   */
  private DSSInvocationContext createTestInvocationContext() {
    return new DSSInvocationContext(getCitizenHsuidHeader(), new Security(), new Header());
  }

  /**
   * @return A valid {@link HsuidHeader} with default information, where the user is
   * a citizen
   */
  private static HsuidHeader getCitizenHsuidHeader() {
    return getValidCitizenHsuidHeader("issuer", "0123456789", "System owner", "System name", "0.0", "Responsible org");
  }

  /**
   * @return A valid {@link HsuidHeader} filled in with the supplied information,
   * where the user is a citizen
   */
  private static HsuidHeader getValidCitizenHsuidHeader(
    String issuer, String cpr, String systemOwner, String systemName, String systemVersion, String responsibleOrg) {
    List<Attribute> attributes = new ArrayList<Attribute>(6);
    addBasicAttributes(true, cpr, systemOwner, systemName, systemVersion, responsibleOrg, attributes);
    return HealthcareServiceUserIdentificationHeaderUtil.createHealthcareServiceUserIdentificationHeader(issuer,
      attributes);
  }

  private static void addBasicAttributes(
    boolean isCitizen, String cpr, String systemOwner, String systemName, String systemVersion, String responsibleOrg,
    List<Attribute> attributes) {
    attributes.add(new UserTypeAttribute(isCitizen));
    attributes.add(new ActingUserCivilRegistrationNumberAttribute("0987654321"));
    attributes.add(new CitizenCivilRegistrationNumberAttribute(cpr));
    attributes.add(new SystemVendorNameAttribute(systemOwner));
    attributes.add(new SystemNameAttribute(systemName));
    attributes.add(new SystemVersionAttribute(systemVersion));
    attributes.add(new OperationsOrganisationNameAttribute(responsibleOrg));
  }

  @Produces
  private Config createTestConfig() {
    TestConfig config = new TestConfig();
    config.set(SXAConfigKeys.PASSWORD.name(), "password");
    config.set(SXAConfigKeys.REQUEST_TIMEOUT_SECONDS.name(), "1");
    config.set(SXAConfigKeys.ROOT_ID.name(), "11.22.33");
    config.set(SXAConfigKeys.SAMTYKKE_KODE.name(), "samtykkekode");
    config.set(SXAConfigKeys.SAMTYKKE_TEKST.name(), "samtykketekst");
    config.set(SXAConfigKeys.SLUTBRUGER_AUTORISATIONSKODE.name(), "slutautorisationskode");
    config.set(SXAConfigKeys.SLUTBRUGER_CPR.name(), "slutcpr");
    config.set(SXAConfigKeys.SLUTBRUGER_EFTERNAVN.name(), "slutefternavn");
    config.set(SXAConfigKeys.SLUTBRUGER_EMAIL.name(), "slutemail");
    config.set(SXAConfigKeys.SLUTBRUGER_FORNAVN.name(), "slutfornavn");
    config.set(SXAConfigKeys.SLUTBRUGER_IDENTIFIKATION.name(), "slutidentifikation");
    config.set(SXAConfigKeys.SLUTBRUGER_ORGANISATION.name(), "slutorganisation");
    config.set(SXAConfigKeys.SLUTBRUGER_ROLLE.name(), "slutrolle");
    config.set(SXAConfigKeys.SLUTBRUGER_STILLING.name(), "slutstilling");
    return config;
  }
}
