/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.util;

import static junit.framework.Assert.assertEquals;
import static junit.framework.Assert.assertFalse;
import static junit.framework.Assert.assertNotNull;
import static junit.framework.Assert.assertTrue;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import javax.xml.ws.Holder;

import org.joda.time.DateTime;
import org.junit.Test;
import org.oasis_open.docs.wss._2004._01.oasis_200401_wss_wssecurity_secext_1_0.Security;

import com.cgi.svareksponering.Relationship;
import com.cgi.svareksponering.Speciale;
import com.cgi.svareksponering.UserRole;

import dk.medcom.dgws._2006._04.dgws_1_0.Header;
import dk.nsi.dds.projects.sxa.common.DocumentIdEncoderDecoder;
import dk.nsi.dds.projects.sxa.common.InvalidSXAQueryParameters;
import dk.nsi.dds.projects.sxa.common.MappedCodedValue;
import dk.nsi.dds.projects.sxa.common.QueryDetails;
import dk.nsi.dds.projects.sxa.common.TestDataHelper;
import dk.nsi.documentsharing.core.metadata.model.CodedValue;
import dk.nsi.documentsharing.docsharingservices.context.DSSInvocationContext;
import dk.nsi.documentsharing.id.IdImpl;
import dk.nsi.documentsharing.xds.impl.ValueConverter;
import dk.nsi.hsuid.CitizenUserRelationAttribute;
import dk.nsi.hsuid.CitizenUserRelationAttribute.LegalValues;
import dk.nsi.hsuid._2016._08.hsuid_1_1.HsuidHeader;
import dk.nsi.hsuid.consumer.HealthcareProfessionalHsuid;
import dk.nsi.hsuid.consumer.impl.HsuidConsumerImpl;

public class SXAInvocationAttributesTest {
  private static DateTime FROM = ValueConverter.convertFromDTM("20150101081500");
  private static DateTime TO = ValueConverter.convertFromDTM("20151201081500");
  private static String PATIENT_ID = "0123456789";
  private static String USER_ID = "6789678967";
  private static TestDataHelper configTestData = new TestDataHelper();
  private static CodedValue pathologyCode = configTestData.practiceSettingCodeForPathology;
  private static CodedValue microbiologyCode = configTestData.practiceSettingCodeForMicrobiology;
  private static List<MappedCodedValue> validPracticeSettingCodesList = configTestData.getMappedPracticeSettingCodes();
  private TestConfig config = new TestConfig();

  @Test
  public void invocationAttributesFromQuery() throws Exception {
    final DSSInvocationContext invocationContext = createInvocationContextForCitizenSelf();
    final QueryDetails queryDetails = createTestQueryDetails(PATIENT_ID, FROM, TO, Arrays.asList(microbiologyCode, pathologyCode));
    SXAInvocationAttributes attributes = SXAInvocationAttributesFactory
      .createInvocationAttributes(invocationContext, queryDetails, config);
    assertEquals(FROM, attributes.getFraDato());
    assertEquals(TO, attributes.getTilDato());
    assertEquals(PATIENT_ID, attributes.getPatientCpr());
    assertNotNull(attributes.getSpecialer());
    assertEquals(2, attributes.getSpecialer().length);
    assertTrue(Arrays.asList(attributes.getSpecialer())
      .contains(Speciale.fromValue(configTestData.mappedPracticeSettingCodeForPathology.getMappedValue())));
    assertTrue(Arrays.asList(attributes.getSpecialer())
      .contains(Speciale.fromValue(configTestData.mappedPracticeSettingCodeForMicrobiology.getMappedValue())));
    assertFalse(Arrays.asList(attributes.getSpecialer())
      .contains(Speciale.fromValue(configTestData.mappedPracticeSettingCodeForBiochemistry.getMappedValue())));
  }

  @Test
  public void invocationAttributesFromNoSpecialties() throws Exception {
    final DSSInvocationContext invocationContext = createInvocationContextForCitizenSelf();
    final QueryDetails emptySpecialtiesQueryDetails = createTestQueryDetails(Collections.emptyList());
    SXAInvocationAttributes attributes = SXAInvocationAttributesFactory
      .createInvocationAttributes(invocationContext, emptySpecialtiesQueryDetails, config);
    assertNotNull(attributes.getSpecialer());
    assertEquals(3, attributes.getSpecialer().length);
    assertTrue(Arrays.asList(attributes.getSpecialer())
      .contains(Speciale.fromValue(configTestData.mappedPracticeSettingCodeForPathology.getMappedValue())));
    assertTrue(Arrays.asList(attributes.getSpecialer())
      .contains(Speciale.fromValue(configTestData.mappedPracticeSettingCodeForMicrobiology.getMappedValue())));
    assertTrue(Arrays.asList(attributes.getSpecialer())
      .contains(Speciale.fromValue(configTestData.mappedPracticeSettingCodeForBiochemistry.getMappedValue())));
  }

  @Test
  public void invocationAttributesForCitizenSelf() throws Exception {
    final DSSInvocationContext invocationContext = createInvocationContextForCitizenSelf();
    SXAInvocationAttributes attributes = SXAInvocationAttributesFactory
      .createInvocationAttributes(invocationContext, createTestQueryDetails(configTestData.getPracticeSettingCodes()), config);
    assertEquals(UserRole.BORGER, attributes.getBrugerType());
    assertEquals(Relationship.SELF, attributes.getRelation());
  }

  @Test
  public void invocationAttributesForParent() throws Exception {
    final DSSInvocationContext invocationContext = createInvocationContext(USER_ID, PATIENT_ID, true, LegalValues.CHILD_CUSTODY_HOLDER);
    SXAInvocationAttributes attributes = SXAInvocationAttributesFactory
      .createInvocationAttributes(invocationContext, createTestQueryDetails(configTestData.getPracticeSettingCodes()), config);
    assertEquals(UserRole.BORGER, attributes.getBrugerType());
    assertEquals(Relationship.PARENT, attributes.getRelation());
  }

  @Test
  public void invocationAttributesForGuardian() throws Exception {
    final DSSInvocationContext invocationContext = createInvocationContext(USER_ID, PATIENT_ID, true, LegalValues.GUARDIAN);
    SXAInvocationAttributes attributes = SXAInvocationAttributesFactory
      .createInvocationAttributes(invocationContext, createTestQueryDetails(configTestData.getPracticeSettingCodes()), config);
    assertEquals(UserRole.BORGER, attributes.getBrugerType());
    assertEquals(Relationship.GUARDIAN, attributes.getRelation());
  }

  private QueryDetails createTestQueryDetails(List<CodedValue> practiceSettingCodesList) throws InvalidSXAQueryParameters {
    return createTestQueryDetails(PATIENT_ID, FROM, TO, practiceSettingCodesList);
  }

  private static QueryDetails createTestQueryDetails(String patientId, DateTime from, DateTime to, List<CodedValue> practiceSettingCodesList) throws InvalidSXAQueryParameters {
    String encoded = new DocumentIdEncoderDecoder(patientId, from, to, practiceSettingCodesList,
      validPracticeSettingCodesList).encode();
    assertTrue(encoded.length() > 0);
    assertTrue(encoded.length() <= 47);
    return new QueryDetails(new IdImpl(encoded), validPracticeSettingCodesList);
  }

  private DSSInvocationContext createInvocationContextForCitizenSelf() {
    return createInvocationContext(PATIENT_ID, PATIENT_ID, true, null);
  }

  private DSSInvocationContext createInvocationContext(
    String userId,
    String citizenId,
    boolean aCitizen, 
    CitizenUserRelationAttribute.LegalValues relation) {
    return new DSSInvocationContext(getCitizenHsuidHeader(userId, citizenId, aCitizen, relation), new Security(), new Header());
  }

  private static HsuidHeader getCitizenHsuidHeader(String userId, String citizenId, boolean aCitizen, LegalValues relation) {
    HsuidConsumerImpl hsuidConsumer = new HsuidConsumerImpl("System owner", "System name", "0.0", "Responsible org", "issuer");
    if (aCitizen) {
      return hsuidConsumer.getAsHeader(hsuidConsumer.createCitizenUser(userId, citizenId, relation));
    } else {
      HealthcareProfessionalHsuid professional = hsuidConsumer.createHealthcareProfessionalUser(userId, "someAuthId");
      professional.addOrganisationUsingIdBySor("someSor");
      professional.addPatientId(citizenId);
      return hsuidConsumer.getAsHeader(professional);
    }
  }

}
