/**
 * The MIT License
 *
 * Original work sponsored and donated by National Board of e-Health (NSI), Denmark
 * (http://www.nsi.dk)
 *
 * Copyright (C) 2016 National Board of e-Health (NSI), Denmark (http://www.nsi.dk)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in the
 * Software without restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 * Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
package dk.nsi.dds.projects.sxa.documentprovider.impl;

import java.util.List;

import javax.ejb.Stateless;
import javax.inject.Inject;

import org.apache.log4j.Logger;

import dk.nsi.dds.projects.sxa.common.InvalidSXAQueryParameters;
import dk.nsi.dds.projects.sxa.common.MappedCodedValue;
import dk.nsi.dds.projects.sxa.common.PracticeSettingCodesMappingConfiguration;
import dk.nsi.dds.projects.sxa.common.QueryDetails;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallException;
import dk.nsi.dds.projects.sxa.documentprovider.exceptions.InvokerCallTimedoutException;
import dk.nsi.dds.projects.sxa.documentprovider.invoker.SXAInvokeHandler;
import dk.nsi.dds.projects.sxa.documentprovider.util.NoMatchingPracticeSettingCodesException;
import dk.nsi.dds.projects.sxa.documentprovider.util.SXAInvocationAttributes;
import dk.nsi.dds.projects.sxa.documentprovider.util.SXAInvocationAttributesFactory;
import dk.nsi.documentsharing.core.metadata.SimpleDocumentMetadata;
import dk.nsi.documentsharing.core.metadata.model.Id;
import dk.nsi.documentsharing.core.retrieve.DocumentsToRetrieve;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocument;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocumentImpl;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocumentsResult;
import dk.nsi.documentsharing.core.retrieve.RetrievedDocumentsResultImpl;
import dk.nsi.documentsharing.docsharingservices.context.DSSInvocationContext;
import dk.nsi.documentsharing.id.RootedId;
import dk.nsi.documentsharing.provider.DocumentRetriever;
import dk.nsi.documentsharing.provider.impl.ProviderConfiguration;
import dk.nsi.documentsharing.provider.impl.ProviderConfigurationHandler;
import dk.nsi.documentsharing.util.perflogger.PerformanceLogger;
import dk.nsi.documentsharing.util.perflogger.PerformanceLogger.Monitor;
import dk.nsi.documentsharing.util.xmlhandling.XStreamFileException;
import dk.nsi.privacy.piw.Constants;
import dk.nsi.util.Config;

@Stateless
public class DocumentRetrieverImpl implements DocumentRetriever {
  private static final Logger logger = Logger.getLogger(DocumentRetrieverImpl.class);
  @Inject
  private SXAInvokeHandler sxaInvokeHandler;
  @Inject
  private ProviderConfigurationHandler providerConfigurationHandler;
  @Inject
  private PracticeSettingCodesMappingConfiguration practiceSettingCodesMappingConfiguration;
  @Inject
  private Config config;

  @Override
  public RetrievedDocumentsResult retrieveDocuments(
    DocumentsToRetrieve documentsToRetrieve, 
    DSSInvocationContext invocationContext) {
    try (Monitor monitor = new PerformanceLogger.Monitor("DocumentRetrieverImpl.retrieveDocuments")) {
      RetrievedDocumentsResultImpl result = new RetrievedDocumentsResultImpl();
      if (documentsToRetrieve == null) {
        logger.error("DocumentsToRetrieve is null");
        return handleError(result);
      }
      if (documentsToRetrieve.getDocumentMetadataList().isEmpty()) {
        logger.error("DocumentsToRetrieve is empty");
        return handleError(result);
      }
      String patientId = documentsToRetrieve.getPatientId().getValue();
      List<MappedCodedValue> validPracticeSettingCodesListPracticeSettingCodesList = practiceSettingCodesMappingConfiguration
        .getPracticeSettingCodesConfiguration().getPracticeSettingCodes();
      for (SimpleDocumentMetadata documentMetadata : documentsToRetrieve.getDocumentMetadataList()) {
        try {
          QueryDetails query = new QueryDetails(getDocumentUniqueIdWithoutRoot(documentMetadata.getDocumentUniqueId()),
            validPracticeSettingCodesListPracticeSettingCodesList);
          if (patientId == null || !patientId.equals(query.getPatientId())) {
            result.addErrorRepository(documentMetadata);
            continue;
          }
          SXAInvocationAttributes attributes = SXAInvocationAttributesFactory
            .createInvocationAttributes(invocationContext, query, config);
          result.add(getDocument(documentMetadata, query, attributes, invocationContext));
        }
        catch (NoMatchingPracticeSettingCodesException e) {
          logger.warn(e.getMessage(), e);
          result.addErrorRepository("Forespørgslens søgeparametre for practiceSettingCode kunne ikke omsættes til laboratorie-specialer", documentMetadata);
        } catch (InvalidSXAQueryParameters e) {
          logger.warn(e.getMessage(), e);
          result.addErrorRepository("Der var en fejl i forespørgslens parametre: " + e.getMessage(), documentMetadata);
        } catch (InvokerCallTimedoutException e) {
          logger.warn(e.getMessage(), e);
          result.addErrorBusyUnreachable(documentMetadata);
        } catch (InvokerCallException e) {
          logger.warn(e.getMessage());
          result.addErrorRepository("Der opstod en fejl under invokeringen af laboratoriesvarsservicen (" + e.getMessage() + ")", documentMetadata);
        } catch (Exception e) {
          logger.warn(e.getMessage(), e);
          result.addErrorRepository("Intern fejl i Svareksponeringsservice XDS-adapter", documentMetadata);
        }
      }
      result.recalculateStatus();
      return result;
    } catch (Exception ex) {
      logger.error(ex.getMessage(), ex);
      return handleError(new RetrievedDocumentsResultImpl());
    }
  }

  private static RetrievedDocumentsResultImpl handleError(RetrievedDocumentsResultImpl result) {
    result.addErrorRepository();
    result.recalculateStatus();
    return result;
  }

  @Override
  public ProviderConfiguration getProviderConfiguration() {
    try {
      return providerConfigurationHandler.readConfiguration();
    } catch (XStreamFileException e) {
      logger.error(e);
      return null;
    }
  }

  private Id getDocumentUniqueIdWithoutRoot(Id idWithRoot) {
    String root = getProviderConfiguration().getUniqueIdRoot();
    if (root == null || root.isEmpty()) {
      throw new IllegalArgumentException("Differing documentUniqueId roots");
    }
    return RootedId.createRootedId(root, idWithRoot.getId()).getExtensionAsId();
  }

  private RetrievedDocument getDocument(
    SimpleDocumentMetadata documentMetadata, 
    QueryDetails query, 
    SXAInvocationAttributes attributes, 
    DSSInvocationContext invocationContext)
      throws InvokerCallTimedoutException, InvokerCallException {
    return new RetrievedDocumentImpl(documentMetadata, Constants.MIME_TYPE,
      sxaInvokeHandler.getDocument(documentMetadata.getDocumentUniqueId().getId(), attributes, invocationContext));
  }
}
